#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

# Contains Python code used for the approach presented in our 'hybrid modeling' paper
# Preprint available at: https://doi.org/10.1101/2022.06.07.495081
# Code available at: https://github.com/Protein-Engineering-Framework/Hybrid_Model

from __future__ import annotations

import os
import pickle
from os import listdir
from os.path import isfile, join
from typing import Union
import logging
logger = logging.getLogger('pypef.dca.hybrid_model')

import numpy as np
import sklearn.base
from scipy.stats import spearmanr
from sklearn.linear_model import Ridge
from sklearn.model_selection import GridSearchCV, train_test_split
from scipy.optimize import differential_evolution

from pypef.utils.variant_data import (
    get_sequences_from_file, get_seqs_from_var_name,
    remove_nan_encoded_positions, get_wt_sequence, split_variants
)

from pypef.dca.plmc_encoding import PLMC, get_dca_data_parallel, get_encoded_sequence, EffectiveSiteError
from pypef.utils.to_file import predictions_out
from pypef.utils.plot import plot_y_true_vs_y_pred
import pypef.dca.gremlin_inference
from pypef.dca.gremlin_inference import GREMLIN


class DCAHybridModel:
    alphas = np.logspace(-6, 6, 100)  # Grid for the parameter 'alpha'.
    parameter_range = [(0, 1), (0, 1)]  # Parameter range of 'beta_1' and 'beta_2' with lb <= x <= ub
    # TODO: Implementation of other regression techniques (CVRegression models)

    def __init__(
            self,
            alphas=alphas,
            parameter_range=None,
            x_train: np.ndarray = None,
            y_train: np.ndarray = None,
            x_test: np.ndarray = None,  # not necessary for training
            y_test: np.ndarray = None,  # not necessary for training
            x_wt=None
    ):
        if parameter_range is None:
            parameter_range = parameter_range
        self._alphas = alphas
        self._parameter_range = parameter_range
        self.x_train = x_train
        self.y_train = y_train
        self.x_test = x_test
        self.y_test = y_test
        self.X = np.concatenate((x_train, x_test), axis=0) if self.x_test is not None else self.x_train
        self.y = np.concatenate((y_train, y_test), axis=0) if self.y_test is not None else self.y_train
        self.x_wild_type = x_wt
        self._spearmanr_dca = self._spearmanr_dca()
        self.beta_1, self.beta_2, self.regressor = self.settings(self.x_train, self.y_train)

    @staticmethod
    def spearmanr(
            y1: np.ndarray,
            y2: np.ndarray
    ) -> float:
        """
        Parameters
        ----------
        y1 : np.ndarray
            Array of target fitness values.
        y2 : np.ndarray
            Array of predicted fitness values.

        Returns
        -------
        Spearman's rank correlation coefficient.
        """
        return spearmanr(y1, y2)[0]

    @staticmethod
    def _standardize(
            x: np.ndarray,
            axis=0
    ) -> np.ndarray:
        """
        Standardizes the input array x by subtracting the mean
        and dividing it by the (sample) standard deviation.

        Parameters
        ----------
        x : np.ndarray
            Array to be standardized.
        axis : integer (default=0)
            Axis to exectute operations on.

        Returns
        -------
        Standardized version of 'x'.
        """
        return np.subtract(x, np.mean(x, axis=axis)) / np.std(x, axis=axis, ddof=1)

    def _delta_x(
            self,
            x: np.ndarray
    ) -> np.ndarray:
        """
        Substracts for each variant the encoded wild-type sequence
        from its encoded sequence.
        
        Parameters
        ----------
        x : np.ndarray
            Array of encoded variant sequences (matrix X).

        Returns
        -------
        Array of encoded variant sequences with substracted encoded
        wild-type sequence.
        """
        return np.subtract(x, self.x_wild_type)

    def _delta_e(
            self,
            x: np.ndarray
    ) -> np.ndarray:
        """
        Calculates the difference of the statistical energy 'dE'
        of the variant and wild-type sequence.

        dE = E (variant) - E (wild-type)
        with E = \sum_{i} h_i (o_i) + \sum_{i<j} J_{ij} (o_i, o_j)

        Parameters
        ----------
        X : np.ndarray
            Array of the encoded variant sequences.

        Returns
        -------
        Difference of the statistical energy between variant 
        and wild-type.
        """
        return np.sum(self._delta_x(x), axis=1)

    def _spearmanr_dca(self) -> float:
        """
        Returns
        -------
        Spearman's rank correlation coefficient of the full
        data and the statistical DCA predictions (difference
        of statistical energies). Used to adjust the sign
        of hybrid predictions, i.e.
            beta_1 * y_dca + beta_2 * y_ridge
        or
            beta_1 * y_dca - beta_2 * y_ridge.
        """
        y_dca = self._delta_e(self.X)
        return self.spearmanr(self.y, y_dca)

    def ridge_predictor(
            self,
            x_train: np.ndarray,
            y_train: np.ndarray,
    ) -> object:
        """
        Sets the parameter 'alpha' for ridge regression.

        Parameters
        ----------
        x_train : np.ndarray
            Array of the encoded sequences for training.
        y_train : np.ndarray
            Associated fitness values to the sequences present
            in 'x_train'.

        Returns
        -------
        Ridge object trained on 'x_train' and 'y_train' (cv=5)
        with optimized 'alpha'. 
        """
        grid = GridSearchCV(Ridge(), {'alpha': self._alphas}, cv=5)
        grid.fit(x_train, y_train)
        return Ridge(**grid.best_params_).fit(x_train, y_train)

    def _y_hybrid(
            self,
            y_dca: np.ndarray,
            y_ridge: np.ndarray,
            beta_1: float,
            beta_2: float
    ) -> np.ndarray:
        """
        Chooses sign for connecting the parts of the hybrid model.

        Parameters
        ----------
        y_dca : np.ndarray
            Difference of the statistical energies of variants
            and wild-type.
        y_ridge : np.ndarray
            (Ridge) predicted fitness values of the variants.
        b1 : float
            Float between [0,1] coefficient for regulating DCA
            model contribution.
        b2 : float
            Float between [0,1] coefficient for regulating ML
            model contribution.

        Returns
        -------
        The predicted fitness value-representatives of the hybrid
        model.
        """
        # Uncomment lines below to see if correlation between
        # y_true and y_dca is positive or negative:
        # logger.info(f'Positive or negative correlation of (all data) y_true '
        #             f'and y_dca (+/-?): {self._spearmanr_dca:.3f}')
        if self._spearmanr_dca >= 0:
            return beta_1 * y_dca + beta_2 * y_ridge
        else:  # negative correlation
            return beta_1 * y_dca - beta_2 * y_ridge

    def _adjust_betas(
            self,
            y: np.ndarray,
            y_dca: np.ndarray,
            y_ridge: np.ndarray
    ) -> np.ndarray:
        """
        Find parameters that maximize the absolut Spearman rank
        correlation coefficient using differential evolution.

        Parameters
        ----------
        y : np.ndarray
            Array of fitness values.
        y_dca : np.ndarray
            Difference of the statistical energies of variants
            and wild-type.
        y_ridge : np.ndarray
            (Ridge) predicted fitness values of the variants.

        Returns
        -------
        'beta_1' and 'beta_2' that maximize the absolut Spearman rank correlation
        coefficient.
        """
        loss = lambda b: -np.abs(self.spearmanr(y, b[0] * y_dca + b[1] * y_ridge))
        minimizer = differential_evolution(loss, bounds=self.parameter_range, tol=1e-4)
        return minimizer.x

    def settings(
            self,
            x_train: np.ndarray,
            y_train: np.ndarray,
            train_size_fit=0.66,
            random_state=42
    ) -> tuple:
        """
        Get the adjusted parameters 'beta_1', 'beta_2', and the
        tuned regressor of the hybrid model.

        Parameters
        ----------
        x_train : np.ndarray
            Encoded sequences of the variants in the training set.
        y_train : np.ndarray
            Fitness values of the variants in the training set.
        train_size_fit : float [0,1] (default 0.66)
            Fraction to split training set into another
            training and testing set.
        random_state : int (default=224)
            Random state used to split.

        Returns
        -------
        Tuple containing the adjusted parameters 'beta_1' and 'beta_2',
        as well as the tuned regressor of the hybrid model.
        """
        try:
            X_ttrain, X_ttest, y_ttrain, y_ttest = train_test_split(
                x_train, y_train,
                train_size=train_size_fit,
                random_state=random_state
            )

        except ValueError:
            """
            Not enough sequences to construct a sub-training and sub-testing 
            set when splitting the training set.

            Machine learning/adjusting the parameters 'beta_1' and 'beta_2' not 
            possible -> return parameter setting for 'EVmutation' model.
            """
            return 1.0, 0.0, None

        """
        The sub-training set 'y_ttrain' is subjected to a five-fold cross 
        validation. This leads to the constraint that at least two sequences
        need to be in the 20 % of that set in order to allow a ranking. 

        If this is not given -> return parameter setting for 'EVmutation' model.
        """
        y_ttrain_min_cv = int(0.2 * len(y_ttrain))  # 0.2 because of five-fold cross validation (1/5)
        if y_ttrain_min_cv < 2:
            return 1.0, 0.0, None

        y_dca_ttest = self._delta_e(X_ttest)

        ridge = self.ridge_predictor(X_ttrain, y_ttrain)
        y_ridge_ttest = ridge.predict(X_ttest)

        beta1, beta2 = self._adjust_betas(y_ttest, y_dca_ttest, y_ridge_ttest)
        return beta1, beta2, ridge

    def hybrid_prediction(
            self,
            x: np.ndarray,
            reg: object,  # any regression-based estimator (from sklearn)
            beta_1: float,
            beta_2: float
    ) -> np.ndarray:
        """
        Use the regressor 'reg' and the parameters 'beta_1'
        and 'beta_2' for constructing a hybrid model and
        predicting the fitness associates of 'X'.

        Parameters
        ----------
        x : np.ndarray
            Encoded sequences X used for prediction.
        reg : object
            Tuned ridge regressor for the hybrid model.
        beta_1 : float
            Float for regulating EVmutation model contribution.
        beta_2 : float
            Float for regulating Ridge regressor contribution.

        Returns
        -------
        Predicted fitness associates of 'X' using the
        hybrid model.
        """
        y_dca = self._delta_e(x)
        if reg is None:
            y_ridge = np.random.random(len(y_dca))  # in order to suppress error
        else:
            y_ridge = reg.predict(x)
        # adjusting: + or - on all data --> +-beta_1 * y_dca + beta_2 * y_ridge
        return self._y_hybrid(y_dca, y_ridge, beta_1, beta_2)

    def split_performance(
            self,
            train_size: float = 0.8,
            n_runs: int = 10,
            seed: int = 42,
            save_model: bool = False
    ) -> dict:
        """
        Estimates performance of the model.

        Parameters
        ----------
        train_size : int or float (default=0.8)
            Number of samples in the training dataset
            or fraction of full dataset used for training.
        n_runs : int (default=10)
            Number of different splits to perform.
        seed : int (default=42)
            Seed for random generator.
        save_model : bool (default=False)
            If True, model is saved using pickle, else not.

        Returns
        -------
        data : dict
            Contains information about hybrid model parameters
            and performance results.
        """
        data = {}
        np.random.seed(seed)

        for r, random_state in enumerate(np.random.randint(100, size=n_runs)):
            x_train, x_test, y_train, y_test = train_test_split(
                self.X, self.y, train_size=train_size, random_state=random_state)
            beta_1, beta_2, reg = self.settings(x_train, y_train)
            if beta_2 == 0.0:
                alpha = np.nan
            else:
                if save_model:
                    pickle.dumps(reg)
                alpha = reg.alpha
            data.update(
                {f'{len(y_train)}_{r}':
                    {
                        'no_run': r,
                        'n_y_train': len(y_train),
                        'n_y_test': len(y_test),
                        'rnd_state': random_state,
                        'spearman_rho': self.spearmanr(
                            y_test, self.hybrid_prediction(
                                x_test, reg, beta_1, beta_2
                            )
                        ),
                        'beta_1': beta_1,
                        'beta_2': beta_2,
                        'alpha': alpha
                    }
                }
            )

        return data

    def ls_ts_performance(self):
        beta_1, beta_2, reg = self.settings(
            x_train=self.x_train,
            y_train=self.y_train
        )
        spearman_r = self.spearmanr(
            self.y_test,
            self.hybrid_prediction(self.x_test, reg, beta_1, beta_2)
        )
        return spearman_r, reg, beta_1, beta_2

    def train_and_test(
            self,
            train_percent_fit: float = 0.66,
            random_state: int = 42
    ):
        """
        Description
        ----------
        Trains the hybrid model on a relative number of all variants
        and returns the individual model contribution weights beta_1 (DCA)
        and beta_2 (ML) as well as the hyperparameter-tuned regression model,
        e.g. to save all the hybrid model parameters for later loading as
        Pickle file.

        Parameters
        ----------
        train_percent_fit: float (default = 0.66)
            Relative number of variants used for model fitting (not
            hyperparameter validation. Default of 0.66 and overall train
            size of 0.8 means the total size for least squares fitting
            is 0.8 * 0.66 = 0.528, thus for hyperparameter validation
            the size is 0.8 * 0.33 = 0.264 and for testing the size is
            1 - 0.528 - 0.264 = 0.208.
        random_state: int (default = 42)
            Random state for splitting (and reproduction of results).

        Returns
        ----------
        beta_1: float
            DCA model contribution to hybrid model predictions.
        beta_2: float
            ML model contribution to hybrid model predictions.
        reg: object
            sklearn Estimator class, e.g. sklearn.linear_model.Ridge
            fitted and with optimized hyperparameters (e.g. alpha).
        self._spearmanr_dca: float
            To determine, if spearmanr_dca (i.e. DCA correlation) and measured
            fitness values is positive (>= 0) or negative (< 0).
        test_spearman_r : float
            Achieved performance in terms of Spearman's rank correlation
            between measured and predicted test set variant fitness values.
        """
        beta_1, beta_2, reg = self.settings(
            x_train=self.x_train,
            y_train=self.y_train,
            train_size_fit=train_percent_fit,
            random_state=random_state
        )

        if len(self.y_test) > 0:
            test_spearman_r = self.spearmanr(
                self.y_test,
                self.hybrid_prediction(
                    self.x_test, reg, beta_1, beta_2
                )
            )
        else:
            test_spearman_r = None
        return beta_1, beta_2, reg, self._spearmanr_dca, test_spearman_r

    def get_train_sizes(self) -> np.ndarray:
        """
        Generates a list of train sizes to perform low-n with.

        Returns
        -------
        Numpy array of train sizes up to 80% (i.e. 0.8 * N_variants).
        """
        eighty_percent = int(len(self.y) * 0.8)

        train_sizes = np.sort(np.concatenate([
            np.arange(15, 50, 5), np.arange(50, 100, 10),
            np.arange(100, 150, 20), [160, 200, 250, 300, eighty_percent],
            np.arange(400, 1100, 100)
        ]))

        idx_max = np.where(train_sizes >= eighty_percent)[0][0] + 1
        return train_sizes[:idx_max]

    def run(
            self,
            train_sizes: list = None,
            n_runs: int = 10
    ) -> dict:
        """

        Returns
        ----------
        data: dict
            Performances of the split with size of the
            training set = train_size and size of the
            test set = N_variants - train_size.
        """
        data = {}
        for t, train_size in enumerate(train_sizes):
            logger.info(f'{t + 1}/{len(train_sizes)}:{train_size}')
            data.update(self.split_performance(train_size=train_size, n_runs=n_runs))
        return data


"""
Below: Some helper functions that call or are dependent on the DCAHybridModel class.
"""


def check_model_type(model: dict | DCAHybridModel | PLMC | GREMLIN):
    """
    Checks type/instance of model.
    """
    if type(model) == dict:
        try:
            model = model['model']
        except KeyError:
            raise SystemError("Unknown model dictionary taken from Pickle file.")
    if type(model) == pypef.dca.plmc_encoding.PLMC:
        return 'PLMC'
    elif type(model) == pypef.dca.hybrid_model.DCAHybridModel:
        return 'Hybrid'
    elif type(model) == pypef.dca.gremlin_inference.GREMLIN:
        return 'GREMLIN'
    elif isinstance(model, sklearn.base.BaseEstimator):
        raise SystemError("Loaded an sklearn ML model. For pure ML-based modeling the "
                          "\'ml\' flag has to be used instead of the \'hybrid\' flag.")
    else:
        raise SystemError('Unknown model/unknown Pickle file.')


def get_model_path(model: str):
    try:
        if isfile(model):
            model_path = model
        elif isfile(f'Pickles/{model}'):
            model_path = f'Pickles/{model}'
        else:
            raise SystemError("Did not find specified model file.")
        return model_path
    except TypeError:
        raise SystemError("No provided model. "
                          "Specify a model for DCA-based encoding.")


def get_model_and_type(params_file: str, substitution_sep: str = '/'):
    file_path = get_model_path(params_file)
    try:
        with open(file_path, 'rb') as read_pkl_file:
            model = pickle.load(read_pkl_file)
            model_type = check_model_type(model)
    except pickle.UnpicklingError:
        model_type = 'PLMC_Params'

    if model_type == 'PLMC_Params':
        model = PLMC(
            params_file=params_file,
            separator=substitution_sep,
            verbose=False
        )
        model_type = 'PLMC'

    else:  # --> elif model_type in ['PLMC', 'GREMLIN', 'Hybrid']:
        model = model['model']

    return model, model_type


def save_model_to_dict_pickle(
        model: DCAHybridModel | PLMC | GREMLIN,
        model_type: str | None = None,
        beta_1: float | None = None,
        beta_2: float | None = None,
        spearman_r: float | None = None,
        regressor: sklearn.base.BaseEstimator = None
):
    try:
        os.mkdir('Pickles')
    except FileExistsError:
        pass

    if model_type is None:
        model_type = 'MODEL'
    # else:
    #    model_type += '_MODEL'
    logger.info(f'Save model as Pickle file... {model_type}')
    pickle.dump(
        {
            'model': model,
            'model_type': model_type,
            'beta_1': beta_1,
            'beta_2': beta_2,
            'spearman_rho': spearman_r,
            'regressor': regressor
        },
        open(f'Pickles/{model_type}', 'wb')
    )


global_model = None
global_model_type = None


def plmc_or_gremlin_encoding(
        variants,
        sequences,
        ys_true,
        params_file,
        substitution_sep='/',
        threads=1,
        verbose=True,
        use_global_model=False
):
    """
    Decides based on the params file input type which DCA encoding to be performed, i.e.,
    GREMLIN or PLMC.
    If use_global_model==True, to avoid each time pickle model file getting loaded, which
    is quite inefficient when performing directed evolution, i.e., encoding of single
    sequences, a global model is stored at the first evolution step and used in the
    subsequent steps.
    """
    global global_model, global_model_type
    if ys_true is None:
        ys_true = np.zeros(np.shape(sequences))
    if use_global_model:
        if global_model is None:
            global_model, global_model_type = get_model_and_type(params_file, substitution_sep)
            model, model_type = global_model, global_model_type
        else:
            model, model_type = global_model, global_model_type
    else:
        model, model_type = get_model_and_type(params_file, substitution_sep)
    if model_type == 'PLMC':
        xs, x_wt, variants, sequences, ys_true = plmc_encoding(
            model, variants, sequences, ys_true, threads, verbose
        )
    elif model_type == 'GREMLIN':
        if verbose:
            logger.info(f"Following positions are frequent gap positions in the MSA "
                        f"and cannot be considered for effective modeling, i.e., "
                        f"substitutions at these positions are removed as these would be "
                        f"predicted as wild type:\n{[gap + 1 for gap in model.gaps]}.\n"
                        f"Effective positions (N={len(model.v_idx)}) are:\n"
                        f"{[v_pos + 1 for v_pos in model.v_idx]}")
        xs, x_wt, variants, sequences, ys_true = gremlin_encoding(
            model, variants, sequences, ys_true,
            shift_pos=1, substitution_sep=substitution_sep
        )
    else:
        raise SystemError(
            f"Found a {model_type.lower()} model as input. Please train a new "
            f"hybrid model on the provided LS/TS datasets."
        )
    assert len(xs) == len(variants) == len(sequences) == len(ys_true)
    return xs, variants, sequences, ys_true, x_wt, model, model_type


def gremlin_encoding(gremlin: GREMLIN, variants, sequences, ys_true, shift_pos=1, substitution_sep='/'):
    """
     Gets X and x_wt for DCA prediction: delta_Hamiltonian respectively
     delta_E = np.subtract(X, x_wt), with X = encoded sequences of variants.
     Also removes variants, sequences, and y_trues at MSA gap positions.
    """
    variants, sequences, ys_true = np.atleast_1d(variants), np.atleast_1d(sequences), np.atleast_1d(ys_true)
    variants, sequences, ys_true = remove_gap_pos(
        gremlin.gaps, variants, sequences, ys_true,
        shift_pos=shift_pos, substitution_sep=substitution_sep
    )
    try:
        xs = gremlin.get_score(sequences, encode=True)
    except SystemError:
        xs = []
    x_wt = gremlin.get_score(np.atleast_1d(gremlin.wt_seq), encode=True)
    return xs, x_wt, variants, sequences, ys_true


def plmc_encoding(plmc: PLMC, variants, sequences, ys_true, threads=1, verbose=False):
    """
    Gets X and x_wt for DCA prediction: delta_E = np.subtract(X, x_wt),
    with X = encoded sequences of variants.
    Also removes variants, sequences, and y_trues at MSA gap positions.
    """
    target_seq, index = plmc.get_target_seq_and_index()
    wt_name = target_seq[0] + str(index[0]) + target_seq[0]
    if verbose:
        logger.info(f"Using to-self-substitution '{wt_name}' as wild type reference. "
                    f"Encoding variant sequences. This might take some time...")
    x_wt = get_encoded_sequence(wt_name, plmc)
    if threads > 1:
        # Hyperthreading, NaNs are already being removed by the called function
        variants, sequences, xs, ys_true = get_dca_data_parallel(
            variants, sequences, ys_true, plmc, threads, verbose=verbose)
    else:
        x_ = plmc.collect_encoded_sequences(variants)
        # NaNs must still be removed
        xs, variants, sequences, ys_true = remove_nan_encoded_positions(
            x_, variants, sequences, ys_true
        )
    return xs, x_wt, variants, sequences, ys_true


def remove_gap_pos(
        gaps,
        variants,
        sequences,
        fitnesses,
        shift_pos=1,
        substitution_sep='/'
):
    """
    Remove gap postions from input variants, sequences, and fitness values
    based on input gaps (gap positions).
    Note that by default, gap positions are shifted by +1 to match the input
    variant identifiers (e.g., variant A123C is removed if gap pos is 122; (122 += 1).

    Returns
    -----------
    variants_v
        Variants with substitutions at valid sequence positions, i.e., at non-gap positions
    sequences_v
        Sequences of variants with substitutions at valid sequence positions, i.e., at non-gap positions
    fitnesses_v
        Fitness values of variants with substitutions at valid sequence positions, i.e., at non-gap positions
    """
    variants_v, sequences_v, fitnesses_v = [], [], []
    valid = []
    for i, variant in enumerate(variants):
        variant = variant.split(substitution_sep)
        for var in variant:
            if int(var[1:-1]) not in [gap + shift_pos for gap in gaps]:
                if i not in valid:
                    valid.append(i)
                    variants_v.append(variants[i])
                    sequences_v.append(sequences[i])
                    fitnesses_v.append(fitnesses[i])
    return variants_v, sequences_v, fitnesses_v


def get_delta_e_statistical_model(
        x_test: np.ndarray,
        x_wt: np.ndarray
):
    """
    Description
    -----------
    Delta_E means difference in evolutionary energy in plmc terms.
    In other words, this is the delta of the sum of Hamiltonian-encoded
    sequences of local fields and couplings of encoded sequence and wild-type
    sequence in GREMLIN terms.

    Parameters
    -----------
    x_test: np.ndarray [2-dim]
        Encoded sequences to be subtracted by x_wt to compute delta E.
    x_wt: np.ndarray [1-dim]
        Encoded wild-type sequence.

    Returns
    -----------
    delta_e: np.ndarray [1-dim]
        Summed subtracted encoded sequences.

    """
    delta_x = np.subtract(x_test, x_wt)
    delta_e = np.sum(delta_x, axis=1)
    return delta_e


def generate_model_and_save_pkl(
        variants,
        ys_true,
        params_file,
        wt,
        train_percent_fit: float = 0.66,  # percent of all data: 0.8 * 0.66
        test_percent: float = 0.2,
        random_state: int = 42,
        substitution_sep = '/',
        threads=1
):
    """
    Description
    -----------
    Save (Ridge) regression model (trained and with tuned alpha parameter)
    with betas (beta_1 and beta_2) as dictionary-structured pickle file.

    Parameters
    ----------
    test_percent: float
        Percent of DataFrame data used for testing. The remaining data is
        used for training (fitting and validation).
    train_percent_fit: float
        Percent of DataFrame data to train on.
        The remaining data is used for validation.
    random_state: int
        Random seed for splitting in train and test data for reproducing results.

    Returns
    ----------
    ()
        Just saving model parameters as pickle file.
    """
    wt_seq = get_wt_sequence(wt)
    variants_splitted = split_variants(variants, substitution_sep)
    variants, ys_true, sequences = get_seqs_from_var_name(wt_seq, variants_splitted, ys_true)

    xs, variants, sequences, ys_true, x_wt, model, model_type = plmc_or_gremlin_encoding(
        variants, sequences, ys_true, params_file, substitution_sep, threads)

    logger.info(
        f'Train size (fitting): {train_percent_fit * 100:.1f} % of training data '
        f'({((1 - test_percent) * train_percent_fit) * 100:.1f} % of all data)\n'
        f'Train size validation: {(1 - train_percent_fit) * 100:.1f} % of training data '
        f'({((1 - test_percent) * (1 - train_percent_fit)) * 100:.1f} % of all data)\n'
        f'Test size: {test_percent * 100:.1f} % ({test_percent * 100:.1f} % of all data)\n'
        f'Using random state: {random_state}...\n'
    )

    x_train, x_test, y_train, y_test = train_test_split(
        xs, ys_true, test_size=test_percent, random_state=random_state
    )

    hybrid_model = DCAHybridModel(
        x_train=x_train,
        y_train=y_train,
        x_test=x_test,
        y_test=y_test,
        x_wt=x_wt
    )

    beta_1, beta_2, reg, spearman_dca, test_spearman_r = hybrid_model.train_and_test(
        train_percent_fit=train_percent_fit,
        random_state=random_state
    )
    if reg is None:
        alpha_ = 'None'
    else:
        alpha_ = f'{reg.alpha:.3f}'
    logger.info(
        f'Individual model weights and regressor hyperparameters:\n'
        f'Hybrid model individual model contributions:\nBeta1 (DCA): '
        f'{beta_1:.3f}, Beta2 (ML): {beta_2:.3f} ('
        f'regressor: Ridge(alpha={alpha_}))\n'
        f'Test performance: Spearman\'s rho = {test_spearman_r:.3f}'
    )
    try:
        os.mkdir('Pickles')
    except FileExistsError:
        pass
    model_name = f'HYBRID{model_type.lower()}'
    save_model_to_dict_pickle(hybrid_model, model_name, beta_1, beta_2, test_spearman_r, reg)


def performance_ls_ts(
        ls_fasta: str | None,
        ts_fasta: str | None,
        threads: int,
        params_file: str,
        model_pickle_file: str | None = None,
        substitution_sep: str = '/',
        label=False
):
    """
    Description
    -----------
    Computes performance based on a (linear) regression model trained
    on the training set by optimizing model hyperparameters based on
    validation performances on training subsets (default: 5-fold CV)
    and predicting test set entries using the hyperparmeter-tuned model
    to estimate performance for model generalization.

    Parameters
    -----------
    ls_fasta: str
        Fasta-like file with fitness values. Will be read and extracted
        for training the regressor.
    ts_fasta: str
        Fasta-like file with fitness values. Used for computing performance
        of the tuned regressor for test set entries (performance metric of
        measured and predicted fitness values).
    threads: int
        Number of threads to use for parallel computing using Ray.
    params_file: str
        PLMC parameter file (containing evolutionary, i.e. MSA-based local
        and coupling terms.
    model: str
        Model to load for TS prediction.
    separator: str
        Character to split the variant to obtain the single substitutions
        (default='/').

    Returns
    -----------
    None
        Just plots test results (predicted fitness vs. measured fitness)
        using def plot_y_true_vs_y_pred.
    """
    test_sequences, test_variants, y_test = get_sequences_from_file(ts_fasta)

    if ls_fasta is not None and ts_fasta is not None:
        train_sequences, train_variants, y_train = get_sequences_from_file(ls_fasta)
        x_train, train_variants, train_sequences, y_train, x_wt, _, model_type = plmc_or_gremlin_encoding(
            train_variants, train_sequences, y_train, params_file, substitution_sep, threads
        )

        x_test, test_variants, test_sequences, y_test, *_ = plmc_or_gremlin_encoding(
            test_variants, test_sequences, y_test, params_file, substitution_sep, threads, verbose=False
        )

        logger.info(f"\nInitial training set variants: {len(train_sequences)}. "
                    f"Remaining: {len(train_variants)} (after removing "
                    f"substitutions at gap positions).\nInitial test set "
                    f"variants: {len(test_sequences)}. Remaining: {len(test_variants)} "
                    f"(after removing substitutions at gap positions)."
                    )

        hybrid_model = DCAHybridModel(
            x_train=np.array(x_train),
            y_train=np.array(y_train),
            x_test=np.array(x_test),
            y_test=np.array(y_test),
            x_wt=x_wt
        )
        model_name = f'HYBRID{model_type.lower()}'

        spearman_r, reg, beta_1, beta_2 = hybrid_model.ls_ts_performance()
        ys_pred = hybrid_model.hybrid_prediction(np.array(x_test), reg, beta_1, beta_2)

        if reg is None:
            alpha_ = 'None'
        else:
            alpha_ = f'{reg.alpha:.3f}'
        logger.info(
            f'Individual model weights and regressor hyperparameters:\n'
            f'Hybrid model individual model contributions: Beta1 (DCA): '
            f'{beta_1:.3f}, Beta2 (ML): {beta_2:.3f} (regressor: '
            f'Ridge(alpha={alpha_}))\nTesting performance...'
        )

        save_model_to_dict_pickle(hybrid_model, model_name, beta_1, beta_2, spearman_r, reg)

    elif ts_fasta is not None and model_pickle_file is not None and params_file is not None:
        logger.info(f'Taking model from saved model (Pickle file): {model_pickle_file}...')

        model, model_type = get_model_and_type(model_pickle_file)

        if model_type != 'Hybrid':  # same as below in next elif
            x_test, test_variants, test_sequences, y_test, x_wt, *_ = plmc_or_gremlin_encoding(
                test_variants, test_sequences, y_test, model_pickle_file, substitution_sep, threads, False)
            ys_pred = get_delta_e_statistical_model(x_test, x_wt)
        else:  # Hybrid model input requires params from plmc or GREMLIN model
            beta_1, beta_2, reg = model.beta_1, model.beta_2, model.regressor
            x_test, test_variants, test_sequences, y_test, *_ = plmc_or_gremlin_encoding(
                test_variants, test_sequences, y_test, params_file,
                substitution_sep, threads, False
            )
            ys_pred = model.hybrid_prediction(x_test, reg, beta_1, beta_2)

    elif ts_fasta is not None and model_pickle_file is None:  # no LS provided --> statistical modeling / no ML
        logger.info(f'No learning set provided, falling back to statistical DCA model: '
                    f'no adjustments of individual hybrid model parameters (beta_1 and beta_2).')
        test_sequences, test_variants, y_test = get_sequences_from_file(ts_fasta)
        x_test, test_variants, test_sequences, y_test, x_wt, model, model_type = plmc_or_gremlin_encoding(
            test_variants, test_sequences, y_test, params_file, substitution_sep, threads
        )

        logger.info(f"Initial test set variants: {len(test_sequences)}. "
                    f"Remaining: {len(test_variants)} (after removing "
                    f"substitutions at gap positions).")

        ys_pred = get_delta_e_statistical_model(x_test, x_wt)

        save_model_to_dict_pickle(model, model_type, None, None, spearmanr(y_test, ys_pred)[0], None)

    else:
        raise SystemError('No Test Set given for performance estimation.')

    spearman_rho = spearmanr(y_test, ys_pred)
    logger.info(f'Spearman Rho = {spearman_rho[0]:.3f}')

    plot_y_true_vs_y_pred(
        np.array(y_test), np.array(ys_pred), np.array(test_variants), label=label, hybrid=True
    )


def predict_ps(  # also predicting "pmult" dict directories
        prediction_dict: dict,
        threads: int,
        separator: str,
        model_pickle_file: str,
        params_file: str = None,
        prediction_set: str = None,
        negative: bool = False
):
    """
    Description
    -----------
    Predicting the fitness of sequences of a prediction set
    or multiple prediction sets that were exemplary created with
    'pypef mkps' based on single substitutional variant data
    provided in a CSV and the wild type sequence:
        pypef mkps --wt WT_SEQ --input CSV_FILE
        [--drop THRESHOLD] [--drecomb] [--trecomb] [--qarecomb] [--qirecomb]
        [--ddiverse] [--tdiverse] [--qdiverse]

    Parameters
    -----------
    prediction_dict: dict
        Contains arguments which directory to predict, e.g. {'drecomb': True},
        than predicts prediction files that are present in this directory, e.g.
        in directory './Recomb_Double_Split'.
    params_file: str
        PLMC couplings parameter file
    threads: int
        Threads used for parallelization for DCA-based sequence encoding
    separator: str
        Separator of individual substitution of variants, default '/'
    model_pickle_file: str
        Pickle file containing the hybrid model and model parameters in
        a dictionary format
    test_set: str = None
        Test set for prediction and plotting of predictions (contains
        true fitness values of variants).
    prediction_set: str = None
        Prediction set for prediction, does not contain true fitness values.
    figure: str = None
        Plotting the test set predictions and the corresponding true fitness
        values.
    label: bool = False
        If True, plots associated variant names of predicted variants.
    negative: bool = False
        If true, negative defines improved variants having a reduced/negative
        fitness compared to wild type.


    Returns
    -----------
    ()
        Writes sorted predictions to files (for [--drecomb] [--trecomb]
        [--qarecomb] [--qirecomb] [--ddiverse] [--tdiverse] [--qdiverse]
        in the respective created folders).

    """
    logger.info(f'Taking model from saved model (Pickle file): {model_pickle_file}...')

    model, model_type = get_model_and_type(model_pickle_file)

    if model_type == 'PLMC':
        logger.info(f'No hybrid model provided – falling back to a statistical DCA model.')
    elif model_type == 'Hybrid':
        beta_1, beta_2, reg = model.beta_1, model.beta_2, model.regressor
        if reg is None:
            alpha_ = 'None'
        else:
            alpha_ = f'{reg.alpha:.3f}'
        logger.info(
            f'Individual model weights and regressor hyperparameters:\n'
            f'Hybrid model individual model contributions: Beta1 (DCA): {beta_1:.3f}, '
            f'Beta2 (ML): {beta_2:.3f} (regressor: Ridge(alpha={alpha_})).'
        )

    pmult = [
        'Recomb_Double_Split', 'Recomb_Triple_Split', 'Recomb_Quadruple_Split',
        'Recomb_Quintuple_Split', 'Diverse_Double_Split', 'Diverse_Triple_Split',
        'Diverse_Quadruple_Split'
    ]
    if True in prediction_dict.values():
        for ps, path in zip(prediction_dict.values(), pmult):
            if ps:  # if True, run prediction in this directory, e.g. for drecomb
                logger.info(f'Running predictions for variant-sequence files in directory {path}...')
                all_y_v_pred = []
                files = [f for f in listdir(path) if isfile(join(path, f)) if f.endswith('.fasta')]
                for i, file in enumerate(files):  # collect and predict for each file in the directory
                    logger.info(f'Encoding files ({i + 1}/{len(files)}) for prediction...\n')
                    file_path = os.path.join(path, file)
                    sequences, variants, _ = get_sequences_from_file(file_path)
                    if model_type != 'Hybrid':
                        x_test, test_variants, x_wt, *_ = plmc_or_gremlin_encoding(
                            variants, sequences, None, model, threads=threads, verbose=False,
                            substitution_sep=separator)
                        ys_pred = get_delta_e_statistical_model(x_test, x_wt)
                    else:  # Hybrid model input requires params from plmc or GREMLIN model
                        ##encoding_model, encoding_model_type = get_model_and_type(params_file)
                        x_test, test_variants, *_ = plmc_or_gremlin_encoding(
                            variants, sequences, None, params_file,
                            threads=threads, verbose=False, substitution_sep=separator
                        )
                        ys_pred = model.hybrid_prediction(x_test, reg, beta_1, beta_2)
                    for k, y in enumerate(ys_pred):
                        all_y_v_pred.append((ys_pred[k], variants[k]))
                if negative:  # sort by fitness value
                    all_y_v_pred = sorted(all_y_v_pred, key=lambda x: x[0], reverse=False)
                else:
                    all_y_v_pred = sorted(all_y_v_pred, key=lambda x: x[0], reverse=True)
                predictions_out(
                    predictions=all_y_v_pred,
                    model='Hybrid',
                    prediction_set=f'Top{path}',
                    path=path
                )
            else:  # check next task to do, e.g., predicting triple substituted variants, e.g. trecomb
                continue

    elif prediction_set is not None:
        sequences, variants, _ = get_sequences_from_file(prediction_set)
        # NaNs are already being removed by the called function
        if model_type != 'Hybrid':  # statistical DCA model
            xs, variants, _, _, x_wt, *_ = plmc_or_gremlin_encoding(
                variants, sequences, None, params_file,
                threads=threads, verbose=False, substitution_sep=separator)
            ys_pred = get_delta_e_statistical_model(xs, x_wt)
        else:  # Hybrid model input requires params from plmc or GREMLIN model
            xs, variants, *_ = plmc_or_gremlin_encoding(
                variants, sequences, None, params_file,
                threads=threads, verbose=True, substitution_sep=separator
            )
            ys_pred = model.hybrid_prediction(xs, reg, beta_1, beta_2)
        assert len(xs) == len(variants)
        y_v_pred = zip(ys_pred, variants)
        y_v_pred = sorted(y_v_pred, key=lambda x: x[0], reverse=True)
        predictions_out(
            predictions=y_v_pred,
            model='Hybrid',
            prediction_set=f'Top{prediction_set}'
        )


def predict_directed_evolution(
        encoder: str,
        variant: str,
        sequence: str,
        hybrid_model_data_pkl: str
) -> Union[str, list]:
    """
    Perform directed in silico evolution and predict the fitness of a
    (randomly) selected variant using the hybrid model. This function opens
    the stored DCAHybridModel and the model parameters to predict the fitness
    of the variant encoded herein using the PLMC class. If the variant
    cannot be encoded (based on the PLMC params file), returns 'skip'. Else,
    returning the predicted fitness value and the variant name.
    """
    if hybrid_model_data_pkl is not None:
        model, model_type = get_model_and_type(hybrid_model_data_pkl)
    else:
        model_type = 'StatisticalModel'  # any name != 'Hybrid'

    if model_type != 'Hybrid':  # statistical DCA model
        xs, variant, _, _, x_wt, *_ = plmc_or_gremlin_encoding(
            variant, sequence, None, encoder, verbose=False, use_global_model=True)
        if not list(xs):
            return 'skip'
        y_pred = get_delta_e_statistical_model(xs, x_wt)
    else:  # model_type == 'Hybrid': Hybrid model input requires params from PLMC or GREMLIN model
        xs, variant, *_ = plmc_or_gremlin_encoding(
            variant, sequence, None, encoder, verbose=False, use_global_model=True
        )
        if not list(xs):
            return 'skip'
        try:
            y_pred = model.hybrid_prediction(np.atleast_2d(xs), model.regressor, model.beta_1, model.beta_2)[0]
        except ValueError:
            raise SystemError(
                "Probably a different model was used for encoding than for modeling; "
                "e.g. using a HYBRIDgremlin model in combination with parameters taken from a PLMC file."
            )
    y_pred = float(y_pred)

    return [(y_pred, variant[0][1:])]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

import warnings
import numpy as np
from scipy import stats
from sklearn.metrics import mean_squared_error, r2_score


def get_performances(
        y_true: list,
        y_pred: list
) -> tuple[float, float, float, float, float]:
    """
    Description
    -----------
    Gets performance metrics (R^2, RMSE, NRMSE, Pearson's r, Spearman's rho)
    between y_true and y_pred.

    Parameters
    -----------
    y_true: list
        Measured fitness values.
    y_pred: list
        Predicted fitness values.

    Returns
    -----------
    r_squared: float
    rmse: float
    nrmse: float
    pearson_r: float
    spearman_rho: float
    """
    y_true = list(y_true)
    y_pred = list(y_pred)
    r_squared = r2_score(y_true, y_pred)
    rmse = np.sqrt(mean_squared_error(y_true, y_pred))
    stddev = np.std(y_true, ddof=1)
    nrmse = rmse / stddev
    with warnings.catch_warnings():  # catching RunTime warning when there's no variance in an array,
        warnings.simplefilter("ignore")  # e.g. [2, 2, 2, 2] which would mean divide by zero
        pearson_r = np.corrcoef(y_true, y_pred)[0][1]
        spearman_rho = stats.spearmanr(y_true, y_pred)[0]

    return r_squared, rmse, nrmse, pearson_r, spearman_rho


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

import logging
logger = logging.getLogger('pypef.dca.dca_run')
import ray

from pypef.utils.variant_data import read_csv, get_wt_sequence
from pypef.dca.plmc_encoding import save_plmc_dca_encoding_model
from pypef.dca.hybrid_model import get_model_and_type, performance_ls_ts, predict_ps, generate_model_and_save_pkl
from pypef.dca.gremlin_inference import save_gremlin_as_pickle, save_corr_csv, plot_all_corr_mtx
from pypef.utils.low_n_mutation_extrapolation import performance_mutation_extrapolation, low_n


def run_pypef_hybrid_modeling(arguments):
    threads = abs(arguments['--threads']) if arguments['--threads'] is not None else 1
    threads = threads + 1 if threads == 0 else threads
    try:
        _, model_type = get_model_and_type(arguments['--params'], arguments['--mutation_sep'])
    except TypeError:
        model_type = 'undefined'
    except SystemError:
        model_type = 'undefined'
    if model_type in ['GREMLIN', 'DCAHybridModel'] and threads > 1:
        logger.info(f'No (Ray) parallelization for {model_type} model...')
    elif model_type not in ['GREMLIN', 'DCAHybridModel'] and threads > 1:
        ray.init()
        logger.info(f'Using {threads} threads for running...')
    if model_type == 'DCAHybridModel':
        logger.info(
            f"Note that the hybrid model only optimizes model performances in terms of "
            f"Spearman's correlation of measured versus predicted values. Further, the "
            f"hybrid approach uses only Ridge regression for supervised ML-based hybrid "
            f"model contribution. In hybrid modeling, the ranks of predictions are "
            f"important and not the exact predicted value."
        )

    if arguments['--ts']:
        performance_ls_ts(
            ls_fasta=arguments['--ls'],
            ts_fasta=arguments['--ts'],
            threads=threads,
            params_file=arguments['--params'],
            model_pickle_file=arguments['--model'],
            substitution_sep=arguments['--mutation_sep'],
            label=arguments['--label']
        )

    elif arguments['--params'] and arguments['--model']:
        prediction_dict = {}
        prediction_dict.update({
            'drecomb': arguments['--drecomb'],
            'trecomb': arguments['--trecomb'],
            'qarecomb': arguments['--qarecomb'],
            'qirecomb': arguments['--qirecomb'],
            'ddiverse': arguments['--ddiverse'],
            'tdiverse': arguments['--tdiverse'],
            'qdiverse': arguments['--qdiverse']
        })

        predict_ps(
            prediction_dict=prediction_dict,
            params_file=arguments['--params'],
            threads=threads,
            separator=arguments['--mutation_sep'],
            model_pickle_file=arguments['--model'],
            prediction_set=arguments['--ps'],
            negative=arguments['--negative']
        )

    elif arguments['train_and_save']:
        variants, fitnesses, _ = read_csv(arguments['--input'])
        generate_model_and_save_pkl(
            variants=variants,
            ys_true=fitnesses,
            params_file=arguments['--params'],
            wt=arguments['--wt'],
            train_percent_fit=arguments['--fit_size'],
            test_percent=arguments['--test_size'],
            random_state=arguments['--rnd_state'],
            substitution_sep=arguments['--mutation_sep'],
            threads=arguments['--threads']
        )

    elif arguments['low_n'] or arguments['extrapolation']:
        if arguments['low_n']:
            low_n(
                encoded_csv=arguments['--input'],
                hybrid_modeling=arguments['hybrid']
            )
        elif arguments['extrapolation']:
            performance_mutation_extrapolation(
                encoded_csv=arguments['--input'],
                cv_regressor=arguments['--regressor'],
                conc=arguments['--conc'],
                hybrid_modeling=arguments['hybrid']
            )

    elif arguments['param_inference']:
        if arguments['--msa']:
            save_gremlin_as_pickle(
                alignment=arguments['--msa'],
                wt_seq=get_wt_sequence(arguments['--wt']),
                opt_iter=arguments['--opt_iter']
            )
        elif arguments['--params']:
            save_plmc_dca_encoding_model(
                params_file=arguments['--params'],
                substitution_sep=arguments['--mutation_sep']
            )

    elif arguments['save_msa_info']:
        gremlin = save_gremlin_as_pickle(
            alignment=arguments['--msa'],
            wt_seq=get_wt_sequence(arguments['--wt']),
            opt_iter=arguments['--opt_iter']
        )
        save_corr_csv(gremlin)
        plot_all_corr_mtx(gremlin)

    else:
        performance_ls_ts(
            ls_fasta=arguments['--ls'],
            ts_fasta=arguments['--ts'],
            threads=threads,
            params_file=arguments['--params'],
            substitution_sep=arguments['--mutation_sep']
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

from __future__ import annotations
import os
import numpy as np
import pandas as pd


amino_acids = [
    'A', 'C', 'D', 'E', 'F',
    'G', 'H', 'I', 'K', 'L',
    'M', 'N', 'P', 'Q', 'R',
    'S', 'T', 'V', 'W', 'Y'
]


def get_wt_sequence(sequence_fasta):
    """
    Gets wild-type sequence from defined input file (can be pure sequence or fasta style)
    """
    if sequence_fasta is None:
        return None
    wild_type_sequence = ""
    try:
        with open(sequence_fasta, 'r') as sf:
            for lines in sf.readlines():
                if lines.startswith(">"):
                    continue
                lines = ''.join(lines.split())
                wild_type_sequence += lines
    except FileNotFoundError:
        raise FileNotFoundError("Did not find FASTA file. Check/specify input FASTA "
                                "sequence file for getting the wild-type sequence.")
    return wild_type_sequence


def read_models(number):
    """
    reads the models found in the file Model_Results.txt.
    If no model was trained, the .txt file does not exist.
    """
    try:
        ls = ""
        with open('Model_Results.txt', 'r') as file:
            for i, lines in enumerate(file):
                if i == 0:
                    if lines[:6] == 'No FFT':
                        number += 2
                if i <= number + 1:
                    ls += lines
        return ls
    except FileNotFoundError:
        return "No Model_Results.txt found."


def absolute_path_cwd_file(file):
    """
    Get the current working directory
    """
    if file is None:
        return None
    return os.path.join(os.getcwd(), file)


def path_aaidx_txt_path_from_utils(filename):
    """
    returns the relative path to the /AAindex folder from the utils directory,
    e.g. path/to/pypef/utils/../aaidx/AAindex/FAUJ880104.txt.
    """
    modules_path = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(modules_path, '..', 'ml', 'AAindex', f'{filename}.txt')


def get_sequences_from_file(
        fasta: str,
        mult_path: str | None = None
) -> (np.ndarray, np.ndarray, np.ndarray):
    """
    "Get_Sequences" reads (learning and test) .fasta and
    .fasta-like ".fasl" format files and extracts the name,
    the target value and the sequence of the protein.
    Only takes one-liner sequences for correct input.
    See example directory for required fasta file format.
    Make sure every marker (> and ;) is seperated by a
    space ' ' from the value respectively name.
    """
    if mult_path is not None:
        os.chdir(mult_path)

    sequences = []
    values = []
    names_of_mutations = []

    with open(fasta, 'r') as f:
        words = ""
        for line in f:
            if line.startswith('>'):
                if words != "":
                    sequences.append(words)
                words = line.split('>')
                names_of_mutations.append(words[1].strip())
                words = ""

            elif line.startswith('#'):
                pass  # are comments

            elif line.startswith(';'):
                if words != "":
                    sequences.append(words)
                words = line.split(';')
                values.append(float(words[1].strip()))
                words = ""

            else:
                try:
                    words += line.strip()
                except IndexError:
                    raise IndexError("Learning or Validation sets (.fasta) likely "
                                     "have emtpy lines (e.g. at end of file)")
        if words != "":
            sequences.append(words)
    # Check consistency
    if len(values) != 0:
        if len(sequences) != len(values):
            raise SystemError(
                f'Error: Number of sequences does not fit with number of target values! '
                f'Number of sequences: {str(len(sequences))}, Number of target values: {str(len(values))}.'
            )
    if mult_path is not None:
        os.chdir('..')

    return np.array(sequences), np.array(names_of_mutations), np.array(values)


def remove_nan_encoded_positions(
        xs: np.ndarray | list,
        *yss
):
    """
    Removes encoded sequence (x) of sequence list xs when NaNs occur in x.
    Also removes the corresponding fitness value y (f(x) --> y) at position i.
    ys can also be any type of list, e.g. variants or sequences.
    """
    xs = list(xs)
    temp = []
    for ys in yss:
        try:
            ys = list(np.atleast_1d(ys))
            if isinstance(ys, pd.Series):
                temp.append(list(ys))
            elif ys is None:
                if len(yss) == 1:
                    temp = (None,)
                else:
                    temp.append([None])
            else:
                if type(ys) == np.ndarray:
                    if np.array(ys).ndim == 0:
                        temp.append([list(np.atleast_1d(ys).tolist())])
                    else:
                        temp.append(list(np.atleast_1d(ys).tolist()))
                else:
                    temp.append(list(ys))
        except ValueError:
            temp.append(list(ys))
    if temp:
        yss = temp
    if not yss == () and not yss == (None,):
        for i, ys in enumerate(yss):
            assert len(xs) == len(ys), "Number of input sequences to be compared unequal."
            try:
                for j, x in enumerate(xs):
                    if np.shape(np.array(xs, dtype='object'))[1] and np.shape(np.array(ys, dtype='object'))[1]:
                        assert len(xs[j]) == len(ys[j]), "Length of input sequences to be compared unequal."
            except IndexError:
                break
    drop = []
    for i, x in enumerate(xs):
        try:
            if None in x:
                drop.append(i)
        except TypeError:
            raise TypeError(
                "Take lists of lists as input, e.g., for single sequence "
                "[[1, 2, 3, 4]]."
            )
    drop = sorted(drop, reverse=True)
    for idx in drop:
        del xs[idx]
        if not yss == () and not yss == (None,):
            for ys in yss:
                del ys[idx]
    if len(yss) == 1:
        return np.array(xs, dtype='object'), np.array(yss[0])

    return np.array(xs, dtype='object'), *np.array(yss, dtype='object')


def get_basename(filename: str) -> str:
    """
    Description
    -----------
    Extracts and returns the basename of the filename.

    Parameters
    ----------
    filename: str

    Returns
    -------
    str
        os.path.basename (filename) string without filename extension
    """
    return os.path.basename(filename).split('.')[0]


def get_seqs_from_var_name(
        wt_seq,
        substitutions,
        fitness_values
) -> tuple[list, list, list]:
    """
    Similar to function above but just returns sequences

    wt: str
        Wild-type sequence as string
    substitutions: list
        List of substiutuions of a single variant of the format:
            - Single substitution variant, e.g. variant A123C: ['A123C']
            - Higher variants, e.g. variant A123C/D234E/F345G: ['A123C', 'D234E, 'F345G']
            --> Full substitutions list, e.g.: [['A123C'], ['A123C', 'D234E, 'F345G']]
    fitness_values: list
        List of ints/floats of the variant fitness values, e.g. for two variants: [1.4, 0.8]
    """
    variant, values, sequences = [], [], []
    for i, var in enumerate(substitutions):  # var are lists of (single or multiple) substitutions
        temp = list(wt_seq)
        name = ''
        separation = 0
        if var == ['WT']:
            name = 'WT'
        else:
            for single_var in var:  # single entries of substitution list
                position_index = int(str(single_var)[1:-1]) - 1
                new_amino_acid = str(single_var)[-1]
                temp[position_index] = new_amino_acid
                # checking if multiple entries are inside list
                if separation == 0:
                    name += single_var
                else:
                    name += '/' + single_var
                separation += 1
        variant.append(name)
        values.append(fitness_values[i])
        sequences.append(''.join(temp))

    return variant, values, sequences


def split_variants(variants, sep='/'):
    """
    Splits variants according to mutation separator.
    """
    variants_splitted = []
    for variant in variants:
        variants_splitted.append(variant.split(sep))
    return variants_splitted


def read_csv(
        file_name: str,
        fitness_key: str = None
) -> tuple[list, list, list]:
    """
    Description
    -----------
    Reads input CSV file and return variants names and
    associated fitness values.

    Parameters
    ----------
    file_name: str
        Name of CSV file to read.
    fitness_key: str
        Name of column containing the fitness values.
        If None, column 1 (0-indexed) will be taken.

    Returns
    -------
    variants: np.ndarray
        Array of variant names
    fitnesses:
        Array of fitness values
    """
    df = pd.read_csv(file_name, sep=';', comment='#')
    if df.shape[1] == 1:
        df = pd.read_csv(file_name, sep=',', comment='#')
    if df.shape[1] == 1:
        df = pd.read_csv(file_name, sep='\t', comment='#')
    if fitness_key is not None:
        fitnesses = df[fitness_key].to_numpy(dtype=float)
    else:
        fitnesses = list(df.iloc[:, 1].to_numpy(dtype=float))
    variants = list(df.iloc[:, 0].to_numpy(dtype=str))
    features = list(df.iloc[:, 2:].to_numpy(dtype=float))

    return variants, fitnesses, features


def generate_dataframe_and_save_csv(
        variants: list,
        sequence_encodings: list,
        fitnesses: list,
        csv_file: str,
        encoding_type: str = '',
        save_df_as_csv: bool = True
) -> pd.DataFrame:
    """
    Description
    -----------
    Creates a pandas.DataFrame from the input data (numpy array including
    variant names, fitnesses, and encoded sequences).
    Writes pandas.DataFrame to a specified CSV file follwing the scheme:
    variants; fitness values; encoded sequences

    Parameters
    ----------
    variants: list
        Variant names.
    fitnesses: list
        Sequence-associated fitness value.
    sequence_encodings: list
        Sequence encodings (feature matrix) of sequences.
    csv_file : str
        Name of the csv file containing variant names and associated fitness values.
    encoding_type: str = ''
        Defines name for saved CSV file based on the chosen encoding technique:
        'aaidx', 'onehot', or 'dca'.
    save_df_as_csv : bool
        Writing DataFrame (Substitution;Fitness;Encoding_Features) to CSV (False/True).

    Returns
    -------
    df_dca: pandas.DataFrame
        Dataframe with variant names, fitness values, and features (encoded sequences).
        If save_df_as_csv is True also writes DF to CSV.
    """
    x = np.stack(sequence_encodings)
    feature_dict = {}            # Collecting features for each MSA position i
    for i in range(x.shape[1]):  # (encoding at pos. i) in a dict
        feature_dict[f'X{i + 1:d}'] = x[:, i]

    df_dca = pd.DataFrame()
    df_dca.insert(0, 'variant', variants)
    df_dca.insert(1, 'y', fitnesses)
    df_dca = pd.concat([df_dca, pd.DataFrame(feature_dict)], axis=1)

    if save_df_as_csv:
        filename = f'{get_basename(csv_file)}_{encoding_type}_encoded.csv'
        df_dca.to_csv(filename, sep=';', index=False)

    return df_dca


def process_df_encoding(df_encoding) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Extracts the array of names, encoded sequences, and fitness values
    of the variants from the dataframe 'self.df_encoding'.
    It is mandatory that 'df_encoding' contains the names of the
    variants in the first column, the associated fitness value in the
    second column, and the encoded sequence starting from the third
    column.

    Returns
    -------
    Tuple of variant names, encoded sequences, and fitness values.
    """
    return (
        df_encoding.iloc[:, 0].to_numpy(),
        df_encoding.iloc[:, 2:].to_numpy(),
        df_encoding.iloc[:, 1].to_numpy()
    )


def read_csv_and_shift_pos_ints(
        infile: str,
        offset: int = 0,
        col_sep: str = ';',
        substitution_sep: str = '/',
        target_column: int = 1
):
    """
    Shifts position of substitutions of variants for all variants in the provided
    CSV file and saves the position-shifted variants with the corresponding fitness
    values to a new CSV file.
    """
    df = pd.read_csv(infile, sep=col_sep, comment='#')
    if df.shape[1] == 1:
        df = pd.read_csv(infile, sep=',', comment='#')
    if df.shape[1] == 1:
        df = pd.read_csv(infile, sep='\t', comment='#')
    try:
        df = df.dropna(subset=df.columns[[target_column]])  # if specific column has a NaN drop entire row
    except IndexError:
        raise IndexError("Did only detect a single column which might indicate a missing "
                         "target value column / a wrong specification of the CSV column "
                         "spearator character (e.g., --sep \';\').")

    column_1 = df.iloc[:, 0]
    column_2 = df.iloc[:, target_column].to_numpy()

    new_col = []

    for variant in column_1:
        if substitution_sep in variant:
            split_vars_list = []
            splitted_var = variant.split(substitution_sep)
            for s_var in splitted_var:
                new_var_int = int(s_var[1:-1]) - offset
                new_variant = s_var[0] + str(new_var_int) + s_var[-1]
                split_vars_list.append(new_variant)
            new_variant = ''
            for i, v in enumerate(split_vars_list):
                if i != len(split_vars_list) - 1:
                    new_variant += f'{v}/'  # substitution_sep replaced by '/'
                else:
                    new_variant += v
            new_col.append(new_variant)
        else:
            new_var_int = int(variant[1:-1]) - offset
            new_variant = variant[0] + str(new_var_int) + variant[-1]
            new_col.append(new_variant)

    data = np.array([new_col, column_2]).T
    new_df = pd.DataFrame(data, columns=['variant', 'fitness'])
    new_df.to_csv(infile[:-4] + '_new' + infile[-4:], sep=';', index=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Validation modules from regression.py for AAindex-based encoding
modified for parallelization of the 566 AAindices used for encoding
with Ray, see https://docs.ray.io/en/latest/index.html.
"""

import matplotlib
matplotlib.use('Agg')
import os
import ray
import warnings

from pypef.utils.variant_data import get_sequences_from_file
from pypef.ml.regression import (
    full_aaidx_txt_path, path_aaindex_dir, AAIndexEncoding, get_regressor_performances
)

# to handle UserWarning for PLS n_components as error and general regression module warnings
warnings.filterwarnings(action='ignore', category=RuntimeWarning, module='sklearn')
warnings.filterwarnings(action='ignore', category=UserWarning, module='sklearn')


@ray.remote
def parallel(
        d,
        core,
        aa_indices,
        train_set,
        test_set,
        regressor='pls',
        no_fft=False
):
    """
    Parallelization of running using the user-defined number of cores.
    Defining the task for each core.
    """
    warnings.filterwarnings("ignore", category=UserWarning)
    warnings.filterwarnings("ignore", category=RuntimeWarning)

    performances = []
    for i in range(d[core][0], d[core][1]):
        aaindex = aa_indices[i]  # Parallelization of AAindex iteration
        sequences_train, _, y_train = get_sequences_from_file(train_set)
        sequences_test, _, y_test = get_sequences_from_file(test_set)
        x_aaidx_train = AAIndexEncoding(full_aaidx_txt_path(aaindex), sequences_train)
        x_aaidx_test = AAIndexEncoding(full_aaidx_txt_path(aaindex), sequences_test)

        if not no_fft:  # X is FFT-ed of encoded alphabetical sequence
            x_train, _ = x_aaidx_train.collect_encoded_sequences()
            x_test, _ = x_aaidx_test.collect_encoded_sequences()
        else:  # X is raw encoded of alphabetical sequence
            _, x_train = x_aaidx_train.collect_encoded_sequences()
            _, x_test = x_aaidx_test.collect_encoded_sequences()

        # If x_learn (or y_learn) is an empty array, the sequence could not be encoded,
        # because of NoneType value. -> Skip
        if x_train == 'skip' or x_test == 'skip':
            continue  # skip the rest and do next iteration
        r2, rmse, nrmse, pearson_r, spearman_rho, regressor, best_params = \
            get_regressor_performances(x_train, x_test, y_train, y_test, regressor)
        if r2 is not None:  # None for metrics if MSE can't be calculated
            performances.append([
                aaindex, r2, rmse, nrmse, pearson_r, spearman_rho, regressor, best_params
            ])

    return performances


def aaindex_performance_parallel(
        train_set,
        test_set,
        cores,
        regressor='pls',
        no_fft=False,
        sort='1'
):
    """
    Parallelization of running using the user-defined number of cores.
    Calling function Parallel to execute the parallel running and
    getting the results from each core each being defined by a result ID.
    """
    aa_indices = [file for file in os.listdir(path_aaindex_dir()) if file.endswith('.txt')]

    split = int(len(aa_indices) / cores)
    last_split = int(len(aa_indices) % cores) + split

    d = {}
    for i in range(cores - 1):
        d[i] = [i * split, i * split + split]

    d[cores - 1] = [(cores - 1) * split, (cores - 1) * split + last_split]

    result_ids = []
    for j in range(cores):  # Parallel running
        result_ids.append(parallel.remote(
            d, j, aa_indices, train_set, test_set, regressor, no_fft))

    results = ray.get(result_ids)

    performances = []
    for core in range(cores):
        for j, _ in enumerate(results[core]):
            performances.append(results[core][j])

    try:
        sort = int(sort)
        if sort == 2 or sort == 3:
            performances.sort(key=lambda x: x[sort])
        else:
            performances.sort(key=lambda x: x[sort], reverse=True)

    except ValueError:
        raise ValueError(
            "Choose between options 1 to 5 (R2, RMSE, NRMSE, "
            "Pearson's r, Spearman's rho."
        )

    return performances



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

import os
import numpy as np


def predictions_out(
        predictions,
        model,
        prediction_set,
        path: str = ''
):
    """
    Writes predictions (of the new sequence space) to text file(s).
    """
    name, value = [], []
    for (val, nam) in predictions:
        name.append(nam)
        value.append('{:f}'.format(val))

    data = np.array([name, value]).T
    col_width = max(len(str(value)) for row in data for value in row) + 5

    head = ['Name', 'Prediction']
    path_ = os.path.join(path, 'Predictions_' + str(model) + '_' + str(prediction_set.split('.')[0]) + '.txt')
    with open(path_, 'w') as f:
        f.write("".join(caption.ljust(col_width) for caption in head) + '\n')
        f.write(len(head)*col_width*'-' + '\n')
        for row in data:
            f.write("".join(str(value).ljust(col_width) for value in row) + '\n')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Modules for making prediction files
from input CSV file (with value separators sep=',' or sep=';')
having the CSV-format

HEADER_VARIANTS;HEADER_VARIANTS_FITNESS
VARIANT_1;FITNESS_VALUE_1
VARIANT_2;FITNESS_VALUE_2
...

according to the self devised prediction set convention
> VARIANT_NAME_1
VARIANT_SEQUENCE_1
> VARIANT_NAME_2
VARIANT_SEQUENCE_2
...
"""


import os
import numpy as np
from tqdm import tqdm


def make_fasta_ps(
        filename,
        wt,
        substitutions
):
    """
    Creates prediction sets (.fasta style files, i.e. without fitness values)
    """
    myfile = open(filename, 'w')
    count = 0
    for i, var in enumerate(substitutions):
        temporary = list(wt)
        name = ''
        separation = 0
        for single_var in var:
            position_index = int(str(single_var)[1:-1]) - 1
            new_amino_acid = str(single_var)[-1]
            temporary[position_index] = new_amino_acid
            if separation == 0:
                name += single_var
            else:
                name += '/' + single_var
            separation += 1
        print('>', name, file=myfile)
        print(''.join(temporary), file=myfile)
        count += 1
    myfile.close()


def make_recombinations_double(arr: tuple) -> list:
    """
    Description
    -----------
    Make double recombinant variants.

    Parameters
    ----------
    arr : tuple
        Lists if single substitutions in tuple, e.g.,
        (['L215F'], ['A217N'], ['R219S'], ['L249Y'])

    Returns
    -------
    doubles : list
        List of double substitution lists, e.g.,
        [['L215F', 'A217N'], ['L215F', 'R219S'], ['L215F', 'L249Y'],
        ['A217N', 'R219S'], ['A217N', 'L249Y'], ['R219S', 'L249Y']]
    """
    doubles = []
    arr_pos = [int(substitution[0][1:-1]) for substitution in arr]
    arr_pos, arr = zip(*sorted(zip(arr_pos, arr), key=lambda x: x[0]))
    for i in tqdm(range(len(arr))):
        for j in range(len(arr)):
            if j > i:
                if (arr[i][0])[1:-1] != (arr[j][0])[1:-1]:
                    doubles.append([arr[i][0], arr[j][0]])
                    if len(doubles) >= 8E04:
                        yield doubles
                        doubles = []
    yield doubles


def make_recombinations_triple(arr: list):
    """
    Description
    -----------
    Make triple recombinant variants.

    Parameters
    ----------
    arr: list
        List of single substitutions in tuple, e.g.,
        (['L215F'], ['A217N'], ['R219S'], ['L249Y'])

    Returns
    -------
    triples: list
        List of triple substitution lists, e.g.,
        [['L215F', 'A217N', 'R219S'], ['L215F', 'A217N', 'L249Y'],
        ['L215F', 'R219S', 'L249Y'], ['A217N', 'R219S', 'L249Y']]
    """
    length = len(arr)
    arr_pos = [int(substitution[0][1:-1]) for substitution in arr]
    arr_pos, arr = zip(*sorted(zip(arr_pos, arr), key=lambda x: x[0]))
    triples = []
    for i in tqdm(range(length)):
        for j in range(length):
            for k in range(length):
                if k > j > i:
                    if (arr[i][0])[1:-1] != (arr[j][0])[1:-1] and \
                            (arr[i][0])[1:-1] != (arr[k][0])[1:-1] and \
                            (arr[j][0])[1:-1] != (arr[k][0])[1:-1]:
                        triples.append([arr[i][0], arr[j][0], arr[k][0]])
                        if len(triples) >= 8E04:
                            yield triples
                            triples = []
    yield triples


def make_recombinations_quadruple(arr):
    """
    Description
    -----------
    Make quadruple recombination variants.

    Parameters
    ----------
    arr: list
        List of single substitutions in tuple, e.g.,
        (['L215F'], ['A217N'], ['R219S'], ['L249Y'])

    Returns
    -------
    quadruples: list
        List of quadruple substitution lists, e.g.,
        [['L215F', 'A217N', 'R219S', 'L249Y']]
    """
    length = len(arr)
    arr_pos = [int(substitution[0][1:-1]) for substitution in arr]
    arr_pos, arr = zip(*sorted(zip(arr_pos, arr), key=lambda x: x[0]))
    quadruples = []
    for i in tqdm(range(length)):
        for j in range(length):
            for k in range(length):
                for l in range(length):
                    if l > k > j > i:
                        if (arr[i][0])[1:-1] != (arr[j][0])[1:-1] and \
                                (arr[i][0])[1:-1] != (arr[k][0])[1:-1] and \
                                (arr[i][0])[1:-1] != (arr[l][0])[1:-1] and \
                                (arr[j][0])[1:-1] != (arr[k][0])[1:-1] and \
                                (arr[j][0])[1:-1] != (arr[l][0])[1:-1] and \
                                (arr[k][0])[1:-1] != (arr[l][0])[1:-1]:
                            quadruples.append([arr[i][0], arr[j][0], arr[k][0], arr[l][0]])
                            if len(quadruples) >= 8E04:
                                yield quadruples
                                quadruples = []
    yield quadruples


def make_recombinations_quintuple(arr):
    """
    Make quintuple recombination variants.

    :parameter arr: List(s) of all available single substitution(s)
    in tuple, e.g.,
    (['L215F'], ['A217N'], ['R219S'], ['L249Y'], ['P252I'])

    :returns quintuples: List of quintuple substitution lists, e.g.,
    [['L215F', 'A217N', 'R219S', 'L249Y', 'P252I']]
    """
    length = len(arr)
    arr_pos = [int(substitution[0][1:-1]) for substitution in arr]
    arr_pos, arr = zip(*sorted(zip(arr_pos, arr), key=lambda x: x[0]))
    quintuples = []
    for i in tqdm(range(length)):
        for j in range(length):
            for k in range(length):
                for l in range(length):
                    for m in range(length):
                        if m > l > k > j > i:
                            if (arr[i][0])[1:-1] != (arr[j][0])[1:-1] and \
                                    (arr[i][0])[1:-1] != (arr[k][0])[1:-1] and \
                                    (arr[i][0])[1:-1] != (arr[l][0])[1:-1] and \
                                    (arr[i][0])[1:-1] != (arr[m][0])[1:-1] and \
                                    (arr[j][0])[1:-1] != (arr[k][0])[1:-1] and \
                                    (arr[j][0])[1:-1] != (arr[l][0])[1:-1] and \
                                    (arr[j][0])[1:-1] != (arr[m][0])[1:-1] and \
                                    (arr[k][0])[1:-1] != (arr[l][0])[1:-1] and \
                                    (arr[k][0])[1:-1] != (arr[m][0])[1:-1] and \
                                    (arr[l][0])[1:-1] != (arr[m][0])[1:-1]:
                                quintuples.append([arr[i][0], arr[j][0], arr[k][0], arr[l][0], arr[m][0]])
                                if len(quintuples) >= 8E04:
                                    yield quintuples
                                    quintuples = []
    yield quintuples


def make_directory_and_enter(directory):
    """
    Makes directory for recombined or diverse prediction sets
    """
    previous_working_directory = os.getcwd()
    try:
        if not os.path.exists(os.path.dirname(directory)):
            os.mkdir(directory)
    except OSError:
        pass
    os.chdir(directory)

    return previous_working_directory


def create_split_files(
        array,
        single_variants,
        wt_sequence,
        name,
        no
):
    """
    Creates split files from given variants for yielded recombined or diverse variants.
    """
    if len(array) > 0:
        number_of_split_files = len(array) / (len(single_variants) * 20 ** 3)
        number_of_split_files = round(number_of_split_files)
        if number_of_split_files == 0:
            number_of_split_files += 1
        split = np.array_split(array, number_of_split_files)
        pwd = make_directory_and_enter(name + '_Split')
        for i in split:
            name_ = name + '_Split' + str(no) + '.fasta'
            make_fasta_ps(name_, wt_sequence, i)

        os.chdir(pwd)

        return ()


def make_ssm_singles(wt_seq, aminoacids):
    """
    Making diverse single-saturation mutagenesis dataset, i.e., all
    19 amino acid substitutions at each wild-type sequence position.
    """
    ssm_singles = []
    for i, aa_wt in enumerate(wt_seq):
        for aa in aminoacids:
            if aa_wt != aa:
                ssm_singles.append([f'{aa_wt}{i+1}{aa}'])
    return ssm_singles


def make_combinations_double_all_diverse(arr, aminoacids):
    """
    Make double substituted naturally diverse variants

    :parameter arr: List of single substitutions in tuple, e.g.,
    (['L215F'], ['A217N'], ['R219S'], ['L249Y'])
    :parameter aminoacids: List of amino acids to combine, e.g., all 20 naturally occuring:
    ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L',
    'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']

    :returns doubles: List of double (fully) diverse substitution tuples, e.g.,
    [('L215A', 'A217C'), ('L215A', 'A217D'), ('L215A', 'A217E'), ('L215A', 'A217F'), ...,
    ('R219Y', 'L249T'), ('R219Y', 'L249V'), ('R219Y', 'L249W'), ('R219Y', 'L249Y')]
    """
    doubles = []
    for i in tqdm(range(len(arr))):
        for j in range(i + 1, len(arr)):
            for k in aminoacids:
                for l in aminoacids:
                    """
                    Make sure that following substitution types are not
                    included for prediction. Examples:
                    1. Both simultaneous substitutions define exactly the 
                      same substitution at the same position, e.g., A1C/A1C:
                        (arr[i][0])[1:-1] != (arr[j][0])[1:-1]
                    2. "To-Wild-Type-Substitutions" at a single position e.g., A1A:
                        ((arr[i][0])[:-1] + k)[0] != ((arr[i][0])[:-1] + k)[-1]  # e.g., A1A
                        ((arr[j][0])[:-1] + l)[0] != ((arr[j][0])[:-1] + l)[-1]  # e.g., C2C
                    3. Just reversed substitution patterns, e.g., A1C/A2D and A2D/A1C
                      in doubles tuple (only possible until results not emptied/yielded 
                      and should generally not occur often):
                        not tuple([(arr[j][0])[:-1] + l, (arr[i][0])[:-1] + k]) in doubles  
                    """
                    if (arr[i][0])[1:-1] != (arr[j][0])[1:-1] and \
                            ((arr[i][0])[:-1] + k)[0] != ((arr[i][0])[:-1] + k)[-1] and \
                            ((arr[j][0])[:-1] + l)[0] != ((arr[j][0])[:-1] + l)[-1] and \
                            not tuple([(arr[j][0])[:-1] + l, (arr[i][0])[:-1] + k]) in doubles:
                        doubles.append(tuple([(arr[i][0])[:-1] + k, (arr[j][0])[:-1] + l]))  # tuple needed for
                        if len(doubles) >= 8E04:                                             # list(dict()):
                            doubles = list(dict.fromkeys(doubles))  # removes duplicated list entries
                            yield doubles
                            doubles = []
    doubles = list(dict.fromkeys(doubles))
    yield doubles


def make_combinations_double_all_diverse_and_all_positions(wt_seq, aminoacids):
    """
    Make double substituted naturally diverse variants

    :parameter arr: List of single substitutions in tuple, e.g.,
    (['L215F'], ['A217N'], ['R219S'], ['L249Y'])
    :parameter aminoacids: List of amino acids to combine, e.g., all 20 naturally occuring:
    ['A', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'K', 'L',
    'M', 'N', 'P', 'Q', 'R', 'S', 'T', 'V', 'W', 'Y']

    :returns doubles: List of double (fully) diverse substitution lists, e.g.,
    [('L215A', 'A217C'), ('L215A', 'A217D'), ('L215A', 'A217E'), ('L215A', 'A217F'), ...,
    ('R219Y', 'L249T'), ('R219Y', 'L249V'), ('R219Y', 'L249W'), ('R219Y', 'L249Y')]
    """
    counter = 0
    doubles = []
    for i in tqdm(range(len(wt_seq))):
        for j in range(i + 1, len(wt_seq)):
            for k in aminoacids:
                pos_1 = wt_seq[i] + str(i + 1) + str(k)
                for l in aminoacids:
                    pos_2 = wt_seq[j] + str(j + 1) + str(l)
                    if pos_1[0] != pos_1[-1] \
                            and pos_2[0] != pos_2[-1] \
                            and pos_1[1:-1] != pos_2[1:-1]:
                        doubles.append(tuple([pos_1, pos_2]))  # tuple needed for
                        if len(doubles) >= 8E04:                    # list(dict()):
                            doubles = list(dict.fromkeys(doubles))  # removes duplicated list entries
                            counter += len(doubles)
                            yield doubles
                            doubles = []
    doubles = list(dict.fromkeys(doubles))
    yield doubles


def make_combinations_triple_all_diverse(arr, aminoacids):
    """
    Make triple substituted naturally diverse variants.
    Analogous to function "make_combinations_double_all_diverse"
    but yielding three combined substitutions.
    """
    triples = []
    for i in tqdm(range(len(arr))):
        for j in range(i + 1, len(arr)):
            for k in range(j + 1, len(arr)):
                for l in aminoacids:
                    for m in aminoacids:
                        for n in aminoacids:
                            if ((arr[i][0])[1:-1]) != ((arr[j][0])[1:-1]) != ((arr[k][0])[1:-1]) and\
                                    ((arr[i][0])[:-1] + l)[0] != ((arr[i][0])[:-1] + l)[-1] and\
                                    ((arr[j][0])[:-1] + m)[0] != ((arr[j][0])[:-1] + m)[-1] and\
                                    ((arr[k][0])[:-1] + n)[0] != ((arr[k][0])[:-1] + n)[-1]:
                                triples.append(tuple([(arr[i][0])[:-1] + l, (arr[j][0])[:-1] + m,
                                                      (arr[k][0])[:-1] + n]))
                                if len(triples) >= 8E04:
                                    triples = list(dict.fromkeys(triples))  # transfer to dict and back to list
                                    yield triples
                                    triples = []
    triples = list(dict.fromkeys(triples))
    yield triples


def make_combinations_quadruple_all_diverse(arr, aminoacids):
    """
    Make quadruple substituted naturally diverse variants.
    Analogous to function "make_combinations_double_all_diverse"
    but yielding four combined substitutions.
    """
    quadruples = []
    for i in tqdm(range(len(arr))):
        for j in range(i + 1, len(arr)):
            for k in range(j + 1, len(arr)):
                for l in range(k + 1, len(arr)):
                    for m in aminoacids:
                        for n in aminoacids:
                            for o in aminoacids:
                                for p in aminoacids:
                                    if ((arr[i][0])[1:-1]) \
                                            != ((arr[j][0])[1:-1]) \
                                            != ((arr[k][0])[1:-1]) \
                                            != ((arr[l][0])[1:-1]) \
                                            and\
                                            ((arr[i][0])[:-1] + m)[0] != ((arr[i][0])[:-1] + m)[-1] and\
                                            ((arr[j][0])[:-1] + n)[0] != ((arr[j][0])[:-1] + n)[-1] and\
                                            ((arr[k][0])[:-1] + o)[0] != ((arr[k][0])[:-1] + o)[-1] and\
                                            ((arr[l][0])[:-1] + p)[0] != ((arr[l][0])[:-1] + p)[-1]:
                                        quadruples.append(tuple([(arr[i][0])[:-1] + m, (arr[j][0])[:-1] + n,
                                                                 (arr[k][0])[:-1] + o, (arr[l][0])[:-1] + p]))
                                        if len(quadruples) >= 8E04:
                                            quadruples = list(dict.fromkeys(quadruples))  # transfer to dict
                                            yield quadruples                              # and back to list
                                            quadruples = []
    quadruples = list(dict.fromkeys(quadruples))
    yield quadruples


if __name__ == '__main__':
    k = list(make_recombinations_quintuple((
        ['A86V'], ['T91S'], ['M108Q'], ['A109E'], ['T111P'], ['A86S'], ['T91E'], ['M108L'], ['A109S'], ['T111G'],
        ['M108R'], ['T111N'], ['T91V'], ['M108T'], ['A109G'], ['T111F'], ['T91A'], ['A109M'], ['A86D'], ['T91R'],
        ['A109K'], ['T111D'], ['T91Q'], ['A109V'], ['T111S'], ['A86C'], ['T91L'], ['A109T'], ['M108S'], ['A109F'],
        ['T111L'], ['A86T'], ['A109Q'], ['M108A'], ['A109P'], ['T111Q'], ['A86N'], ['T91Y'], ['A109L'], ['T111A'],
        ['T91F'], ['A109Y'], ['A86I'], ['A109D'], ['M108K'], ['M108I'], ['T91N'], ['T111C'], ['T91M'], ['T91C'],
        ['M108P'], ['T111M'], ['T91H'], ['M108C'], ['M108F'], ['M108G'], ['A109N'], ['M108E'], ['A109W'], ['M108W'],
        ['A109I'], ['T91P'], ['M108H'], ['T91D'], ['A109R'], ['T91I'], ['M108Y'], ['T91G'], ['T91W'], ['A86R'],
        ['T91K'], ['T111Y'], ['M108D'], ['A86W'], ['M108V'], ['T111I'], ['M108N'], ['A109C'], ['A109H']
    )))
    for i, k_ in enumerate(k):
        print(i + 1, np.shape(k_))
    # (10 * 80,000 (* 5)) + (1 * 2503 (* 5))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Main modules for regression/ML including feature generation
(i.e. sequence encoding), cross-validation-based hyperparameter
tuning, prediction, and plotting routines.
"""


import os
from typing import Union

import logging
logger = logging.getLogger('pypef.ml.regression')
import matplotlib
matplotlib.use('Agg')  # no plt.show(), just save plot
import matplotlib.pyplot as plt
import numpy as np
import pickle
from tqdm import tqdm  # progress bars
from sklearn.model_selection import LeaveOneOut
from sklearn.model_selection import KFold
from sklearn.metrics import mean_squared_error
from sklearn.model_selection import GridSearchCV  # default: refit=True

# import regression models
from sklearn.cross_decomposition import PLSRegression
from sklearn.ensemble import RandomForestRegressor
from sklearn.svm import SVR
from sklearn.neural_network import MLPRegressor
from sklearn.linear_model import Ridge, Lasso, ElasticNet

from pypef.utils.variant_data import (
    amino_acids, get_sequences_from_file, get_basename
)
from pypef.utils.plot import plot_y_true_vs_y_pred
from pypef.utils.performance import get_performances
from pypef.dca.hybrid_model import plmc_or_gremlin_encoding

import warnings
warnings.filterwarnings(action='ignore', category=RuntimeWarning, module='numpy')
# ignoring warnings of PLS regression when using n_components
warnings.filterwarnings(action='ignore', category=RuntimeWarning, module='sklearn')
warnings.filterwarnings(action='ignore', category=UserWarning, module='sklearn')
warnings.filterwarnings(action='ignore', category=DeprecationWarning, module='sklearn')
# FutureWarning: The default of 'normalize' will be set to False in version 1.2 and deprecated in version 1.4.
# If you wish to scale the data, use Pipeline with a StandardScaler in a preprocessing stage.
# sklearn: The default value of 'normalize' should be changed to False in linear models where now normalize=True
warnings.filterwarnings(action='ignore', category=FutureWarning, module='sklearn')


# globals
x_train = None
x_test = None
y_train = None
y_test = None
train_variants = None
test_variants = None
train_sequences = None
test_sequences = None
model_type = None


def read_models(number):
    """
    reads the models found in the file Model_Results.txt.
    If no model was trained, the .txt file does not exist.
    """
    try:
        ls = ""
        with open('Model_Results.txt', 'r') as file:
            for i, lines in enumerate(file):
                if i == 0:
                    if lines[:6] == 'No FFT':
                        number += 2
                if i <= number + 1:
                    ls += lines
        return ls
    except FileNotFoundError:
        return "No Model_Results.txt found."


def full_aaidx_txt_path(filename):
    """
    returns the path of an index inside the folder /AAindex/,
    e.g. /path/to/pypef/ml/AAindex/FAUJ880104.txt.
    """
    modules_path = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(modules_path, 'AAindex', filename)


def path_aaindex_dir():
    """
    returns the absolute path to the /AAindex folder,
    e.g. /path/to/AAindex/.
    """
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), 'AAindex')


class AAIndex:
    """
    gets all the information that are given in each AAindex file.
    For the program routine it provides the library to enable translation
    of the alphabetical amino acid sequence to an array of numericals.
    """
    def __init__(self, filename):
        self.file = filename
        self.accession_number = None
        self.data_description = None
        self.pmid = None
        self.authors = None
        self.title_of_article = None
        self.journal_reference = None

    def general_information(self):
        """
        Gets and allocates general information based on the AAindex file
        format defined by file sections 'H', 'D', 'E', 'A', 'T', 'J'
        """
        with open(self.file, 'r') as f:
            for line in f:
                # try/ except "removes" empty lines.
                try:
                    words = line.split()
                    id_letter = words[0]
                except IndexError:
                    break

                # Extract some general information about AAindex file.
                if id_letter == 'H':
                    self.accession_number = words[1]
                elif id_letter == 'D':
                    self.data_description = words[1]
                elif id_letter == 'E':
                    self.pmid = words[1:]
                elif id_letter == 'A':
                    self.authors = ' '.join(words[1:])
                elif id_letter == 'T':
                    self.title_of_article = ' '.join(words[1:])
                elif id_letter == 'J':
                    self.journal_reference = ' '.join(words[1:])

    def encoding_dictionary(self):
        """
        Get numerical values of AAindex for each amino acid
        """
        try:
            with open(self.file, 'r') as f:
                for line in f:
                    # try/ except "removes" empty lines
                    try:
                        words = line.split()
                        id_letter = words[0]
                    except IndexError:
                        break

                    # Extract numerical values of AAindex.
                    if id_letter == 'I':

                        keys = []
                        for word in words[1:]:
                            keys.append(word[0])
                            keys.append(word[-1])

                        values = []
                        for row in range(2):
                            line = f.readline()
                            strings = line.split()
                            for idx, string in enumerate(strings):
                                # Some amino acids may have no value
                                try:
                                    strings[idx] = float(string)
                                except ValueError:
                                    strings[idx] = None
                            values.append(strings)
                        values = np.reshape(np.array(values).T, len(keys))

                        return dict(zip(keys, values))
        except FileNotFoundError:
            raise FileNotFoundError(
                "Probably you used an encoding technique option in combination with a model "
                "that was created using another encoding option (e.g. pypef ml -e aaidx -m "
                "ONEHOT -p TS.fasta) which is not allowed."
            )


class AAIndexEncoding:
    """
    converts the string sequence into a list of numericals
    using the AAindex translation library; Fourier trans-
    forming the numerical array that was translated by
    get_numerical_sequence --> do_fourier,computing the input
    matrices X and Y for the regressor (get_x_and_y).
    Returns FFT-ed encoded sequences (amplitudes),
    and raw_encoded sequences (raw_numerical_sequences).
    """
    def __init__(
            self,
            aaindex_file=None,
            sequences: list = None,
    ):
        aaidx = AAIndex(aaindex_file)
        self.dictionary = aaidx.encoding_dictionary()
        self.sequences = sequences

    def get_numerical_sequence(self, sequence):
        return np.array([self.dictionary[aminoacid] for aminoacid in sequence])

    @staticmethod
    def do_fourier(sequence):
        """
        This static function does the Fast Fourier Transform.
        Since the condition

            len(Array) = 2^k -> k = log_2(len(Array))  , k in N

        must be satisfied, the array must be reshaped (zero padding)
        if k is no integer value. The verbose parameter prints also
        the real and imaginary part separately.
        """
        threshold = 1e-8  # errors due to computer uncertainties
        k = np.log2(sequence.size)  # get exponent k
        mean = np.mean(sequence, axis=0)  # calculate mean of numerical array
        sequence = np.subtract(sequence, mean)  # subtract mean to avoid artificial effects of FT

        if abs(int(k) - k) > threshold:  # check if length of array fulfills previous equation
            numerical_sequence_reshaped = np.zeros(pow(2, (int(k) + 1)))  # reshape array
            for index, value in enumerate(sequence):
                numerical_sequence_reshaped[index] = value
            sequence = numerical_sequence_reshaped

        fourier_transformed = np.fft.fft(sequence)  # FFT
        ft_real = np.real(fourier_transformed)
        ft_imag = np.imag(fourier_transformed)

        x = np.linspace(1, sequence.size, sequence.size)  # frequencies
        x = x / max(x)  # normalization of frequency

        amplitude = ft_real * ft_real + ft_imag * ft_imag

        if max(amplitude) != 0:
            amplitude = np.true_divide(amplitude, max(amplitude))  # normalization of amplitude

        return amplitude, x

    def aaidx_and_or_fft_encode_sequence(self, sequence):
        """
        getting the input matrices X (FFT amplitudes) and Y (variant labels)
        """
        num = self.get_numerical_sequence(sequence)
        # Numerical sequence gets expended by zeros so that also different
        # lengths of sequences can be processed using '--nofft' option
        k = np.log2(len(num))
        if abs(int(k) - k) > 1e-8:  # check if length of array fulfills previous equation
            raw_numerical_seq = np.append(num, np.zeros(pow(2, (int(k) + 1)) - len(num)))  # reshape array
        else:
            raw_numerical_seq = num

        if None not in num:  # Not all amino acids could be encoded with the corresponding AAindex
            amplitudes_, frequencies_ = self.do_fourier(num)  # --> None values in encoded sequence
        else:  # If None in encoded Sequence, do not further use encoding (and FFT not possible)
            return [None], [None]
        # Fourier spectra are mirrored at frequency = 0.5 -> No more information at higher frequencies
        half = len(frequencies_) // 2  # // for integer division
        amplitude = amplitudes_[:half]   # FFT-ed encoded amino acid sequences
        # Appended zeros of raw encoding allow also prediction of differently sizes sequences

        # return -> X_fft_encoding, X_raw_encoding
        return amplitude, raw_numerical_seq

    def collect_encoded_sequences(self):
        """
        Loop over all sequences to encode each and collect
        and return all encoded sequences
        """
        # There may be amino acids without a value in AAindex
        # Skip these indices
        fft_encoded_sequences, raw_encoded_sequences = [], []
        for sequence in self.sequences:
            fft_encoded_sequence, raw_encoded_sequence = self.aaidx_and_or_fft_encode_sequence(sequence)
            if None in raw_encoded_sequence:
                return 'skip', 'skip'  # skipping this AAindex
            else:
                fft_encoded_sequences.append(fft_encoded_sequence)
                raw_encoded_sequences.append(raw_encoded_sequence)

        return fft_encoded_sequences, raw_encoded_sequences


class OneHotEncoding:
    """
    Generates an one-hot encoding, i.e. represents
    the current amino acid at a position as 1 and
    the other (19) amino acids as 0. Thus, the encoding
    of a sequence has the length 20 x sequence length.
    E.g. 'ACDY' --> [1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                     0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                     0,0,1,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
                     0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1]
    """
    def __init__(
            self,
            sequences: list,
            verbose=True
    ):
        self.sequences = sequences
        self.amino_acids = amino_acids  # imported, 20 standard AAs
        self.verbose = verbose

    def encoding_dict(self) -> dict[str, np.ndarray]:
        encoding_dict = {}
        for idx, amino_acid in enumerate(self.amino_acids):
            encoding_vector = np.zeros(20, dtype=int)
            encoding_vector[idx] = 1
            encoding_dict.update({amino_acid: encoding_vector})
        return encoding_dict

    def one_hot_encode_sequence(self, sequence: str) -> np.ndarray:
        encoded_sequence = []
        for aminoacid in sequence:
            encoded_sequence.append(self.encoding_dict()[aminoacid])
        return np.concatenate(encoded_sequence)

    def collect_encoded_sequences(self, verbose: bool = None) -> np.ndarray:
        if verbose is None:
            disable = not self.verbose
        else:
            disable = not verbose
        if len(np.atleast_1d(self.sequences)) == 1:  # always silence for single (wt) sequence
            disable = True
        encoded_sequences = []
        for sequence in tqdm(self.sequences, disable=disable):
            encoded_sequences.append(self.one_hot_encode_sequence(sequence))
        return np.array(encoded_sequences)


def pls_loocv(
        x_train: np.ndarray,
        y_train: np.ndarray
) -> Union[tuple[str, str], tuple[PLSRegression, dict]]:
    """
    PLS regression with LOOCV n_components tuning as described by Cadet et al.
    https://doi.org/10.1186/s12859-018-2407-8
    https://doi.org/10.1038/s41598-018-35033-y
    Hyperparameter (N component) tuning of PLS regressor, can achieve slightly better
    """
    mean_squared_error_list = []
    for n_comp in range(1, 10):  # n_comp = 1, 2,..., 9
        try:
            pls = PLSRegression(n_components=n_comp)
            loo = LeaveOneOut()
            y_pred_loo = []
            y_test_loo = []
            for train, test in loo.split(x_train):
                x_learn_loo = []
                y_learn_loo = []
                x_test_loo = []
                for j in train:
                    x_learn_loo.append(x_train[j])
                    y_learn_loo.append(y_train[j])
                for k in test:
                    x_test_loo.append(x_train[k])
                    y_test_loo.append(y_train[k])
                x_learn_loo = np.array(x_learn_loo)
                x_test_loo = np.array(x_test_loo)
                y_learn_loo = np.array(y_learn_loo)
                try:
                    pls.fit(x_learn_loo, y_learn_loo)
                except ValueError:  # scipy/linalg/decomp_svd.py ValueError:
                    continue        # illegal value in %dth argument of internal gesdd
                y_pred_loo.append(pls.predict(x_test_loo)[0][0])
        except np.linalg.LinAlgError:  # numpy.linalg.LinAlgError: SVD did not converge
            continue
        try:
            mse = mean_squared_error(y_test_loo, y_pred_loo)
            mean_squared_error_list.append(mse)
        except ValueError:  # MSE could not be calculated (No values due to numpy.linalg.LinAlgErrors)
            return 'skip', 'skip'
    mean_squared_error_list = np.array(mean_squared_error_list)
    idx = np.where(mean_squared_error_list == np.min(mean_squared_error_list))[0][0] + 1
    # Model is fitted with best n_components (lowest MSE)
    best_params = {'n_components': idx}
    regressor_ = PLSRegression(n_components=best_params.get('n_components'))

    return regressor_, best_params


def cv_regression_options(regressor: str) -> GridSearchCV:
    """
    Returns the CVRegressor with the tunable regression-specific hyperparameter grid
    for training a regression model.
    Regression options are
        - Partial Least Squares Regression
        - Random Forest Regression
        - Support Vector Machines Regression
        - Multilayer Perceptron Regression
        - Ridge Regression
        - Lasso Regression
        - ElasticNet Regression
    """
    if regressor == 'pls':
        params = {'n_components': list(np.arange(1, 10))}  # n_comp = 1, 2,..., 9
        regressor_ = GridSearchCV(PLSRegression(), param_grid=params, cv=5)  # iid in future versions redundant

    elif regressor == 'rf':
        params = {  # similar parameter grid as Xu et al., https://doi.org/10.1021/acs.jcim.0c00073
            'random_state': [42],  # state determined
            'n_estimators': [100, 250, 500, 1000],  # number of individual decision trees in the forest
            'max_features': ['auto', 'sqrt', 'log2']  # “auto” -> max_features=n_features,
            # “sqrt” -> max_features=sqrt(n_features) “log2” -> max_features=log2(n_features)
        }
        regressor_ = GridSearchCV(RandomForestRegressor(), param_grid=params, cv=5)

    elif regressor == 'svr':
        params = {  # similar parameter grid as Xu et al.
            'C': [2 ** 0, 2 ** 2, 2 ** 4, 2 ** 6, 2 ** 8, 2 ** 10, 2 ** 12],  # Regularization parameter
            'gamma': [0.1, 0.01, 0.001, 0.0001, 0.00001]  # often 1 / n_features or 1 / (n_features * X.var())
        }
        regressor_ = GridSearchCV(SVR(), param_grid=params, cv=5)

    elif regressor == 'mlp':
        params = {
            # feedforward network trained via backpropagation – here only using a single hidden layer
            'hidden_layer_sizes': [i for i in range(1, 12)],  # size of hidden layer [(1,), (2,), ..., (12,)]
            'activation': ['relu'],  # rectified linear unit
            'solver': ['adam', 'lbfgs'],  # ADAM: A Method for Stochastic Optimization , or Limited-memory BFGS
            'learning_rate': ['constant'],  # learning rate given by ‘learning_rate_init’
            'learning_rate_init': [0.001, 0.01, 0.1],  # only used when solver=’sgd’ or ‘adam’
            'max_iter': [1000, 200],  # for stochastic solvers (‘sgd’, ‘adam’) determines epochs
            'random_state': [42]
        }
        regressor_ = GridSearchCV(MLPRegressor(), param_grid=params, cv=5)

    elif regressor == 'elasticnet' or regressor == 'l1l2':
        # Linear regression with combined L1 and L2 priors as regularizer.
        # min(w):  ||y - Xw||^2_2 + alpha*l1_ratio*||w||_1 + 0.5*alpha*(1 - l1_ratio)*||w||^2_2
        params = {
            'alpha': np.logspace(-6, 6, num=100)  # {1.000E-06, 1.322E-06, 1.748E-06, ..., 1.000E06}
        }
        regressor_ = GridSearchCV(ElasticNet(), param_grid=params, cv=5)

    elif regressor == 'ridge' or regressor == 'l2':
        # Performs L2 regularization, i.e., adds penalty equivalent to square of the magnitude of coefficients
        # Majorly used to prevent overfitting, since it includes all the features
        # min(w): ||y - Xw||^2_2 + alpha*||w||^2_2
        # in case of exorbitantly high features, it will pose computational challenges.
        params = {
            # alpha = 0 is equivalent to an ordinary least square regression
            # higher values of alpha reduce overfitting, significantly high values can
            # cause underfitting as well (e.g., regularization strength alpha = 5)
            'alpha': np.logspace(-6, 6, num=100)  # {1.000E-06, 1.322E-06, 1.748E-06, ..., 1.000E06}
        }
        regressor_ = GridSearchCV(Ridge(), param_grid=params, cv=5)

    elif regressor == 'lasso' or regressor == 'l1':
        # Lasso model fit with Least Angle Regression a.k.a. Lars.
        # Performs L1 regularization, i.e., adds penalty equivalent to absolute value of the magnitude of coefficients
        # min(w): ||y - Xw||^2_2 + alpha*||w||_1
        # Provides sparse solutions: computationally efficient as features with zero coefficients can be ignored
        params = {
            # alpha = 0 is equivalent to an ordinary least square Regression
            'alpha': np.logspace(-6, 6, num=100)  # {1.000E-06, 1.322E-06, 1.748E-06, ..., 1.000E06}
        }
        regressor_ = GridSearchCV(Lasso(), param_grid=params, cv=5)

    else:
        raise SystemError("Did not find specified regression model as valid option. See '--help' for valid "
                          "regression model options.")

    return regressor_


def get_regressor_performances(
        x_learn: list,
        x_test: list,
        y_learn: list,
        y_test: list,
        regressor: str = 'pls',
        verbose: bool = False
):
    """
    The function get_regressor_performances takes features and labels from the
    learning and test set.

    When using 'pls_loocv' as regressor, the MSE is calculated for all LOOCV
    sets for predicted vs true labels (mse = mean_squared_error(y_test_loo, y_pred_loo)
    for a fixed number of components for PLS regression.
    In the next iteration, the number of components is increased by 1 (number_of_components += 1)
    and the MSE is calculated for this regressor. The loop breaks if i > 9.
    Finally, the model of the single AAindex model with the lowest MSE is chosen.

    When using other regressors the parameters are tuned using GridSearchCV.

    This function returnes performance (R2, (N)RMSE, Pearson's r) and model parameters.
    """
    regressor = regressor.lower()
    best_params = None

    if regressor == 'pls_loocv':  # PLS LOOCV tuning
        regressor_, best_params = pls_loocv(x_learn, y_learn)
        if regressor_ == 'skip':
            return [None, None, None, None, None, regressor, None]

    # other regression options (k-fold CV tuning)
    else:
        regressor_ = cv_regression_options(regressor)
    try:
        if verbose:
            logger.info('CV-based training of regression model...')
        regressor_.fit(x_learn, y_learn)  # fit model
    except ValueError:  # scipy/linalg/decomp_svd.py --> ValueError('illegal value in %dth argument of internal gesdd'
        return [None, None, None, None, None, regressor, None]

    if regressor != 'pls_loocv':  # take best parameters for the regressor and the AAindex
        best_params = regressor_.best_params_

    y_pred = []
    try:
        for y_p in regressor_.predict(x_test):  # predict validation entries with fitted model
            y_pred.append(float(y_p))
    except ValueError:
        raise ValueError("Above error message exception indicates that your test set may be empty.")

    r2, rmse, nrmse, pearson_r, spearman_rho = get_performances(y_test, y_pred)

    return r2, rmse, nrmse, pearson_r, spearman_rho, regressor, best_params


def performance_list(
        train_set: str,
        test_set: str,
        encoding: str = 'aaidx',
        regressor: str = 'pls',
        no_fft: bool = False,
        sort: str = '1',
        couplings_file: str = None,
        threads: int = 1  # for parallelization of DCA-based encoding
):
    """
    returns the sorted list of all the model parameters and the
    performance values (R2 etc.) from function get_performances.
    """
    global x_train, y_train, train_variants, train_sequences, \
        x_test, y_test, test_variants, test_sequences, \
        model_type
    encoding = encoding.lower()
    performance_list = []
    train_sequences, train_variants, y_train = get_sequences_from_file(train_set)
    test_sequences, test_variants, y_test = get_sequences_from_file(test_set)
    if encoding == 'onehot':  # OneHot-based encoding
        x_onehot_train = OneHotEncoding(train_sequences)
        x_onehot_test = OneHotEncoding(test_sequences)
        x_train = x_onehot_train.collect_encoded_sequences()
        x_test = x_onehot_test.collect_encoded_sequences()
        r2, rmse, nrmse, pearson_r, spearman_rho, regression_model, \
            params = get_regressor_performances(x_train, x_test, y_train, y_test, regressor, verbose=True)
        if r2 is not None:  # get_regressor_performances() returns None for metrics if MSE can't be calculated
            performance_list.append([
                encoding.upper(), r2, rmse, nrmse, pearson_r,
                spearman_rho, regression_model, params
            ])
    elif encoding == 'dca':  # 'plmc' or encoding == 'gremlin':
        x_train, train_variants, train_sequences, y_train, x_wt, model, model_type = plmc_or_gremlin_encoding(
            train_variants, train_sequences, y_train, couplings_file, threads=threads
        )
        x_test, test_variants, test_sequences, y_test, x_wt, model, model_type = plmc_or_gremlin_encoding(
            test_variants, test_sequences, y_test, couplings_file, threads=threads
        )

        r2, rmse, nrmse, pearson_r, spearman_rho, regression_model, \
            params = get_regressor_performances(x_train, x_test, y_train, y_test, regressor, verbose=True)
        if r2 is not None:  # get_regressor_performances() returns None for metrics if MSE can't be calculated
            performance_list.append([
                encoding.upper(), r2, rmse, nrmse, pearson_r,
                spearman_rho, regression_model, params
            ])

    else:  # AAindex-based encoding
        aa_indices = [file for file in os.listdir(path_aaindex_dir()) if file.endswith('.txt')]
        # loop over the 566 AAindex entries, encode with each AAindex and test performance
        # can be seen as a AAindex hyperparameter search on the test set  --> also see CV performance
        # in created folder across all data to ensure a relatively well generalizable model
        for index, aaindex in enumerate(tqdm(aa_indices)):
            x_aaidx_train = AAIndexEncoding(full_aaidx_txt_path(aaindex), train_sequences)
            if not no_fft:  # X is FFT-ed of encoded alphabetical sequence
                x_train, _ = x_aaidx_train.collect_encoded_sequences()
            else:  # X is raw encoded of alphabetical sequence
                _, x_train = x_aaidx_train.collect_encoded_sequences()
            x_aaidx_test = AAIndexEncoding(full_aaidx_txt_path(aaindex), test_sequences)
            if not no_fft:  # X is FFT-ed of the encoded alphabetical sequence
                x_test, _ = x_aaidx_test.collect_encoded_sequences()
            else:  # X is the raw encoded of alphabetical sequence
                _, x_test = x_aaidx_test.collect_encoded_sequences()
            # If x_learn or x_test contains None, the sequence could not be (fully) encoded --> Skip
            if x_train == 'skip' or x_test == 'skip':
                continue  # skip the rest and do next iteration
            r2, rmse, nrmse, pearson_r, spearman_rho, regression_model, \
                params = get_regressor_performances(x_train, x_test, y_train, y_test, regressor)
            if r2 is not None:  # get_regressor_performances() returns None for metrics if MSE can't be calculated
                performance_list.append([
                    aaindex, r2, rmse, nrmse, pearson_r,
                    spearman_rho, regression_model, params
                ])

    try:
        sort = int(sort)
        if sort == 2 or sort == 3:
            performance_list.sort(key=lambda x: x[sort])
        else:
            performance_list.sort(key=lambda x: x[sort], reverse=True)

    except ValueError:
        raise ValueError("Choose between options 1 to 5 (R2, RMSE, NRMSE, Pearson's r, Spearman's rho.")

    return performance_list


def formatted_output(
        performance_list,
        no_fft=False,
        minimum_r2=-1E9
):
    """
    Takes the sorted list from function r2_list and writes the model names with an R2 ≥ 0
    as well as the corresponding parameters for each model so that the user gets
    a list (Model_Results.txt) of the top ranking models for the given validation set.
    """

    index, value, value2, value3, value4, value5, regression_model, params = [], [], [], [], [], [], [], []

    for (idx, val, val2, val3, val4, val5, r_m, pam) in performance_list:
        if val >= minimum_r2:
            index.append(get_basename(idx))
            value.append('{:f}'.format(val))
            value2.append('{:f}'.format(val2))
            value3.append('{:f}'.format(val3))
            value4.append('{:f}'.format(val4))
            value5.append('{:f}'.format(val5))
            regression_model.append(r_m.upper())
            params.append(pam)

    if len(value) == 0:  # Criterion of not finding suitable model is defined by Minimum_R2
        raise SystemError(f'No model with minimum R2 ({minimum_r2}).')

    data = np.array([index, value, value2, value3, value4, value5, regression_model, params]).T
    col_width = max(len(str(value)) for row in data for value in row[:-1]) + 5

    head = ['Index', 'R2', 'RMSE', 'NRMSE', 'Pearson\'s r', 'Spearman\'s rho', 'Regression', 'Model parameters']
    with open('Model_Results.txt', 'w') as f:
        if no_fft is not False:
            f.write("No FFT used in this model construction, performance"
                    " represents model accuracies on raw encoded sequence data.\n\n")

        heading = "".join(caption.ljust(col_width) for caption in head) + '\n'
        f.write(heading)

        row_length = []
        for row in data:
            row_ = "".join(str(value).ljust(col_width) for value in row) + '\n'
            row_length.append(len(row_))
        row_length_max = max(row_length)
        f.write(row_length_max * '-' + '\n')

        for row in data:
            f.write("".join(str(value).ljust(col_width) for value in row) + '\n')


def cross_validation(
        x: np.ndarray,
        y: np.ndarray,
        regressor_,  # Union[PLSRegression, Ridge, Lasso, ElasticNet, ...]
        n_samples: int = 5):
    """
    Perform k-fold cross-validation on the input data (encoded sequences and
    corresponding fitness values) with default k = 5. Returns all predicted
    fitness values of the length y (e.g. (1/5)*len(y) * 5 = 1*len(y)).
    """
    # perform k-fold cross-validation on all data
    # k = Number of splits, change for changing k in k-fold splitting, default: 5
    y_test_total = []
    y_predicted_total = []

    kf = KFold(n_splits=n_samples, shuffle=True)

    for train_index, test_index in kf.split(y):
        y = np.array(y)
        try:
            x_train, x_test = x[train_index], x[test_index]
            y_train, y_test = y[train_index], y[test_index]

            for numbers in y_test:
                y_test_total.append(numbers)
            regressor_.fit(x_train, y_train)
            y_pred_test = regressor_.predict(x_test)

            for values in y_pred_test:
                y_predicted_total.append(float(values))
        except UserWarning:
            continue

    return y_test_total, y_predicted_total


def get_regressor(
        regressor: str,
        parameter: dict
):
    """
    Returns the tuned CVRegressor with the tuned hyperparameters.
    Regression options are
        - Partial Least Squares Regression
        - Random Forest Regression
        - Support Vector Machines Regression
        - Multilayer Perceptron Regression
        - Ridge Regression
        - Lasso Regression
        - ElasticNet Regression
    """
    if regressor == 'pls' or regressor == 'pls_loocv':
        # n_components according to lowest MSE for validation set
        regressor_ = PLSRegression(n_components=parameter.get('n_components'))

    elif regressor == 'rf':
        regressor_ = RandomForestRegressor(
            random_state=parameter.get('random_state'),
            n_estimators=parameter.get('n_estimators'),
            max_features=parameter.get('max_features')
        )

    elif regressor == 'svr':
        regressor_ = SVR(C=parameter.get('C'), gamma=parameter.get('gamma'))

    elif regressor == 'mlp':
        regressor_ = MLPRegressor(
            hidden_layer_sizes=parameter.get('hidden_layer_sizes'),
            activation=parameter.get('activation'),
            solver=parameter.get('solver'),
            learning_rate=parameter.get('learning_rate'),
            learning_rate_init=parameter.get('learning_rate_init'),
            max_iter=parameter.get('max_iter'),
            random_state=parameter.get('random_state')
        )

    elif regressor == 'ridge' or regressor == 'l2':
        regressor_ = Ridge(
            alpha=parameter.get('alpha')
        )

    elif regressor == 'lasso' or regressor == 'l1':
        regressor_ = Lasso(
            alpha=parameter.get('alpha')
        )

    elif regressor == 'elasticnet' or regressor == 'l1l2':
        regressor_ = ElasticNet(
            alpha=parameter.get('alpha')
        )

    else:
        raise SystemError("Did not find specified regression model as valid option. "
                          "See '--help' for valid regression model options.")

    return regressor_


def encode_based_on_type(
        encoding: str,
        variants,
        sequences,
        y_true=None,
        couplings_file=None,
        idx=None,
        threads=1,
        no_fft=True,
        substitution_sep='/',
        verbose=True
):
    if y_true is None:
        y_true = np.zeros(np.shape(sequences))
    if encoding == 'aaidx':  # AAindex encoding technique
        encoder = AAIndexEncoding(full_aaidx_txt_path(idx), list(np.atleast_1d(sequences)))
        if no_fft is False:  # use FFT on encoded sequences (default)
            x, _ = encoder.collect_encoded_sequences()
        else:  # use raw encoding (no FFT used on encoded sequences)
            _, x = encoder.collect_encoded_sequences()
    elif encoding == 'onehot':  # OneHot encoding technique
        encoder = OneHotEncoding(sequences)
        x = encoder.collect_encoded_sequences()
    elif encoding == 'dca':  # PLMC or GREMLIN-based encoding
        if len(sequences) == 1:
            use_global_model = True
        else:
            use_global_model = False
        x, variants, sequences, y_true, x_wt, model, model_type = plmc_or_gremlin_encoding(
            variants, sequences, y_true, couplings_file, substitution_sep, threads, verbose, use_global_model
        )
    else:
        raise SystemError("Unknown encoding option.")

    assert len(x) == len(variants) == len(sequences) == len(y_true)
    return x, variants, sequences, y_true


def crossval_on_all(x_train, x_test, y_train, y_test, regressor: str, parameter, idx=None, no_fft=False):
    """
    Use e.g. 80 %  of all data for learning (hyperparameter tuning via validation)
    and e.g. 20 % for testing 5 times for 5-fold cross validation.
    """
    name = get_basename(idx)
    cv_filename = os.path.join('CV_performance', f'{name}_{regressor.upper()}_CV_Results.txt')
    try:
        os.remove(cv_filename)
    except FileNotFoundError:
        pass
    file = open(cv_filename, 'w')
    file.write('5-fold cross-validated performance of top '
               'models for validation set across all data.\n\n')
    if no_fft:
        file.write("No FFT used in this model construction, performance represents"
                   " model accuracies on raw encoded sequence data.\n\n")
    file.close()

    x = np.concatenate([x_train, x_test])
    y = np.concatenate([y_train, y_test])

    regressor_ = get_regressor(regressor, parameter)
    # perform 5-fold cross-validation on all data (on X and Y)
    n_samples = 5
    y_test_total, y_predicted_total = cross_validation(x, y, regressor_, n_samples)

    r_squared, rmse, nrmse, pearson_r, spearman_rho = \
        get_performances(y_test_total, y_predicted_total)

    with open(cv_filename, 'a') as f:
        f.write('Regression type: {}; Parameter: {}; Encoding index: {}\n'.format(
            regressor.upper(), parameter, name))
        f.write('R2 = {:.5f}; RMSE = {:.5f}; NRMSE = {:.5f}; Pearson\'s r = {:.5f};'
                ' Spearman\'s rho = {:.5f}\n\n'.format(r_squared, rmse, nrmse, pearson_r, spearman_rho))

    figure, ax = plt.subplots()
    legend = r'$R^2$' + f' = {r_squared:.3f}' + f'\nRMSE = {rmse:.3f}' + f'\nNRMSE = {nrmse:.3f}' + \
             f'\nPearson\'s ' + r'$r$' + f' = {pearson_r:.3f}' + f'\nSpearman\'s ' + \
             fr'$\rho$ = {spearman_rho:.3f}' + '\n' + fr'($N$ = {len(y_test_total)})'
    ax.scatter(
        y_test_total, y_predicted_total,
        marker='o', s=20, linewidths=0.5, edgecolor='black', label=legend, alpha=0.8
    )
    ax.plot([min(y_test_total) - 1, max(y_test_total) + 1],
            [min(y_predicted_total) - 1, max(y_predicted_total) + 1], 'k', lw=0.5)

    ax.set_xlabel('Measured')
    ax.set_ylabel('Predicted')
    ax.legend(prop={'size': 8})
    plt.savefig(
        os.path.join('CV_performance', f'{name}_{regressor.upper()}_{n_samples}-fold-CV.png'),
        dpi=500
    )
    plt.close('all')


def save_model(
        path,
        performances,
        training_set,
        test_set,
        threshold=5,
        encoding='aaidx',
        regressor='pls',
        no_fft=False,
        train_on_all=False,
        couplings_file=None,
        threads: int = 1,
        label=False
):
    """
    Function Save_Model saves the best -s THRESHOLD models as 'Pickle'
    files (pickle.dump), which can be loaded again for doing predictions.
    Also, in Save_Model included is the def cross_validation-based computing
    of the k-fold CV performance of the n component-optimized model on all
    data (learning + test set); by default  k  is 5 (n_samples = 5).
    Plots of the CV performance for the t best models are stored inside the
    folder CV_performance.
    """
    global x_train, y_train, train_variants, train_sequences, \
        x_test, y_test, test_variants, test_sequences, \
        model_type
    logger.info('Encoding and cross validation on all data (creating folder CV_performance)...')
    regressor = regressor.lower()
    try:
        os.mkdir('CV_performance')
    except FileExistsError:
        pass
    try:
        os.mkdir('Pickles')
    except FileExistsError:
        pass
    if encoding != 'aaidx' and x_train is not None and x_test is not None:
        pass  # take global encodings instead of recomputing DCA encodings
    else:
        train_sequences, train_variants, y_train = get_sequences_from_file(training_set)
        test_sequences, test_variants, y_test = get_sequences_from_file(test_set)
    for i, t in enumerate(range(threshold)):
        try:
            idx = performances[t][0]
            parameter = performances[t][7]

            if encoding != 'aaidx' and x_train is not None and x_test is not None:
                pass  # take global encodings instead of recomputing DCA encodings
            else:
                x_train, train_variants, train_sequences, y_train = encode_based_on_type(
                    encoding, train_variants, train_sequences, y_train, couplings_file, idx, threads, no_fft
                )
                x_test, test_variants, test_sequences, y_test = encode_based_on_type(
                    encoding, test_variants, test_sequences, y_test, couplings_file, idx, threads, no_fft
                )

            crossval_on_all(x_train, x_test, y_train, y_test, regressor, parameter, idx, no_fft)
            regressor_ = get_regressor(regressor, parameter)
            if train_on_all:  # Train model hyperparameters based on all available data (training + test set)
                # But, no generalization performance can be estimated as the model also trained on the test set
                x = np.concatenate([x_train, x_test])
                y = np.concatenate([y_train, y_test])
                regressor_.fit(x, y)
            else:
                # fit (only) on full learning set (FFT or noFFT is defined already above)
                regressor_.fit(x_train, y_train)
            # 2D prediction array output to 1D
            y_test_pred = np.array(regressor_.predict(x_test)).flatten()
            plot_y_true_vs_y_pred(
                y_true=y_test,
                y_pred=y_test_pred,
                variants=test_variants,
                label=label,
                hybrid=False,
                name=f'{get_basename(idx)}_{regressor.upper()}_'
            )
            name = get_basename(idx)
            if model_type in ['PLMC', 'GREMLIN'] and encoding not in ['aaidx', 'onehot']:
                name = 'ML' + model_type.lower()
            logger.info(f'Saving model as {name}')
            file = open(os.path.join(path, 'Pickles', name), 'wb')
            pickle.dump(regressor_, file)
            file.close()

        except IndexError:
            raise IndexError
        #    break

        if encoding == 'onehot' or encoding == 'dca':   # only 1 model/encoding  -->
            break                                       # no further iteration needed, thus break loop


def predict(
        path,
        model,
        prediction_set=None,
        encoding='aaidx',
        mult_path=None,
        no_fft=False,
        variants=None,
        sequences=None,
        couplings_file=None,
        threads: int = 1,  # for parallelization of DCA-based encoding
        substitution_sep='/',
        verbose=False
):
    """
    The function Predict is used to perform predictions.
    Saved pickle files of models will be loaded again:
      mod = pickle.load(file)
    and used for predicting the label y (y = mod.predict(x))
    of sequences given in the Prediction_Set.fasta.
    """
    # model defines pickle to load (and thus determines encoding AAidx)
    file = open(os.path.join(path, 'Pickles', str(model)), 'rb')
    loaded_model = pickle.load(file)
    file.close()
    idx = None
    if encoding == 'aaidx':
        idx = model + '.txt'
    if sequences is None and variants is None:  # File-based prediction
        sequences, variants, _ = get_sequences_from_file(prediction_set, mult_path)

    try:
        x, variants, sequences, _ = encode_based_on_type(
            encoding, variants, sequences, None, couplings_file,
            idx, threads, no_fft, substitution_sep, verbose
        )
    except SystemError:
        return 'skip'
    if type(x) == list:
        if not x:
            return 'skip'
    elif type(x) == np.ndarray:
        if not x.any():
            return 'skip'

    assert len(variants) == len(x)

    try:
        ys = loaded_model.predict(x)
    except ValueError:
        raise SystemError(
            "If you used an encoding such as onehot, make sure to use the correct model, e.g. -m ONEHOT. "
            "If you used an AAindex-encoded model you likely tried to predict using a model encoded with "
            "(or without) FFT featurization ('--nofft') while the model was trained without (or with) FFT "
            "featurization so check Model_Results.txt line 1, if the models were trained with or without FFT."
        )
    except AttributeError:
        raise SystemError(
            "The model specified is likely a hybrid or pure statistical DCA (and no pure ML model).\n"
            "Check the specified model provided via the \'-m\' flag."
        )

    predictions = [(float(ys[i]), variants[i]) for i in range(len(ys))]  # List of tuples

    # Pay attention if increased negative values would define a better variant --> use negative flag
    predictions.sort()
    predictions.reverse()  # if predictions array is too large?
    return predictions


def predict_ts(
        path,
        model,
        test_set=None,
        encoding='aaidx',
        idx=None,
        no_fft=False,
        couplings_file=None,
        label=False,
        threads: int = 1  # for parallelization of DCA-based encoding
):
    """
    The function Predict is used to perform predictions.
    Saved pickle files of models will be loaded again:
      mod = pickle.load(file)
    and used for predicting the label y (y = mod.predict(x))
    of sequences given in the Prediction_Set.fasta.
    """
    file = open(os.path.join(path, 'Pickles', str(model)), 'rb')
    loaded_model = pickle.load(file)
    file.close()
    if type(loaded_model) == dict:
        loaded_model = loaded_model['model']
    if encoding == 'aaidx':
        idx = model + '.txt'

    sequences, variants, y_test = get_sequences_from_file(test_set)
    x, variants, sequences, y_test, *_ = encode_based_on_type(
        encoding, variants, sequences, y_test, couplings_file, idx, threads, no_fft
    )
    if type(x) == list:
        if not x:
            return 'skip'
    elif type(x) == np.ndarray:
        if not x.any():
            return 'skip'
    if encoding != 'aaidx':
        idx = encoding

    assert len(variants) == len(x)

    try:
        y_pred = loaded_model.predict(x)
        y_pred = list(np.array(y_pred).flatten())
    except ValueError:
        raise SystemError(
            "If you used an encoding such as onehot, make sure to use the correct model, e.g. -m ONEHOT. "
            "If you used an AAindex-encoded model you likely tried to predict using a model encoded with "
            "(or without) FFT featurization ('--nofft') while the model was trained without (or with) FFT "
            "featurization so check Model_Results.txt line 1, if the models were trained with or without FFT."
        )
    except AttributeError:
        raise SystemError(
            "The model specified is likely a hybrid or pure statistical DCA (and no pure ML model).\n"
            "Check the specified model provided via the \'-m\' flag."
        )

    plot_y_true_vs_y_pred(y_test, y_pred, variants, label, hybrid=False, name=f'{get_basename(idx).upper()}_')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

import numpy as np
from scipy import stats
import matplotlib.pyplot as plt
from adjustText import adjust_text
import logging
logger = logging.getLogger('pypef.ml.regression')

from pypef.utils.performance import get_performances


def plot_y_true_vs_y_pred(
        y_true: np.ndarray,
        y_pred: np.ndarray,
        variants: np.ndarray,  # just required for labeling
        label=False,
        hybrid=False,
        name: str = ''
):
    """
    Plots predicted versus true values using the hybrid model for prediction.
    Function called by function predict_ps.
    """
    figure, ax = plt.subplots()
    if hybrid:
        spearman_rho = stats.spearmanr(y_true, y_pred)[0]
        ax.scatter(y_true, y_pred, marker='o', s=20, linewidths=0.5, edgecolor='black', alpha=0.7,
                   label=f'Spearman\'s ' + fr'$\rho$ = {spearman_rho:.3f}' + '\n' + fr'($N$ = {len(y_true)})')
        file_name = name + 'DCA_Hybrid_Model_Performance.png'
    else:
        r_squared, rmse, nrmse, pearson_r, spearman_rho = get_performances(
            y_true=y_true, y_pred=y_pred
        )
        ax.scatter(
            y_true, y_pred, marker='o', s=20, linewidths=0.5, edgecolor='black', alpha=0.7,
            label=r'$R^2$' + f' = {r_squared:.3f}' + f'\nRMSE = {rmse:.3f}' + f'\nNRMSE = {nrmse:.3f}' +
                  f'\nPearson\'s ' + r'$r$'+f' = {pearson_r:.3f}' + f'\nSpearman\'s ' +
                  fr'$\rho$ = {spearman_rho:.3f}' + '\n' + fr'($N$ = {len(y_true)})'
        )
        file_name = name + 'ML_Model_Performance.png'
        # x = np.linspace(min(y_pred), max(y_pred), 100)
        # ax.plot(x, x, color='black', linewidth=0.25)  # plot diagonal line
    ax.legend(prop={'size': 8})
    ax.set_xlabel('Measured')
    ax.set_ylabel('Predicted')
    logger.info('Plotting...')
    if label:
        logger.info('Adjusting variant labels for plotting can take some '
                    'time (the limit for labeling is 150 data points)...')
        if len(y_true) < 150:
            texts = [ax.text(y_true[i], y_pred[i], txt, fontsize=4)
                     for i, txt in enumerate(variants)]
            adjust_text(
                texts, only_move={'points': 'y', 'text': 'y'}, force_points=0.5, lim=250)
        else:
            logger.info("Terminating label process. Too many variants "
                        "(> 150) for plotting (labels would overlap).")
    # Uncomment for renaming new plots
    # i = 1
    # while os.path.isfile(file_name):
    #     i += 1  # iterate until finding an unused file name
    #     file_name = f'DCA_Hybrid_Model_LS_TS_Performance({i}).png'
    plt.savefig(file_name, dpi=500)
    plt.close('all')



#!/usr/bin/env python3
# for installation run me with: pip install .
#                      or with: pip install -e .


from setuptools import setup, find_packages
from pypef import __version__


with open("requirements.txt", "r", encoding="utf-8") as install_requirements:
    requirements = install_requirements.read()

setup(
    name='pypef',
    version=__version__.split('-')[0],
    author='Niklas Siedhoff & Alexander-Maurice Illig',
    author_email='n.siedhoff@biotec.rwth-aachen.de',
    license='CC BY-NC-SA 4.0',
    description='A command-line interface (CLI) tool for performing data-driven protein engineering '
                'by building machine learning (ML)-trained regression models from sequence variant '
                'fitness data (in CSV format) based on different techniques for protein sequence encoding. '
                'Next to building pure ML models, \'hybrid modeling\' is also possible using a blended '
                'model optimized for predictive contributions of a statistical and an ML-based prediction.',
    long_description='For detailed description including a short Jupyter Notebook-based '
                     'tutorial please refer to the GitHub page.',
    long_description_content_type='text/markdown',
    url='https://github.com/Protein-Engineering-Framework/PyPEF',
    py_modules=['pypef'],
    packages=find_packages(include=['pypef', 'pypef.*']),
    package_data={'pypef': ['ml/AAindex/*', 'ml/AAindex/Refined_cluster_indices_r0.93_r0.97/*']},
    include_package_data=True,
    install_requires=[requirements],
    python_requires='>= 3.9, < 3.12',
    keywords='Pythonic Protein Engineering Framework',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence'
    ],
    entry_points={
        'console_scripts': [
            'pypef = pypef.main:run_main'
        ],
    }
)
